/* * Self-Revocation Free Downgrading - A kernel module that applies information flow
 * control to protect system integrity without self-revocation
 * Copyright (C) 2014 Secure Systems Laboratory, Stony Brook University
 *
 * This file is part of Self-Revocation Free Downgrading.
 *
 * Self-Revocation Free Downgrading is free software: you can redistribute it
 * and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Self-Revocation Free Downgrading is distributed in the hope that it will be
 * useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Self-Revocation Free Downgrading.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

#ifndef __SRFD_LABEL_H__
#define __SRFD_LABEL_H__


/* Label data structure:
 * Consists of three parts-
 * 1. physical label:
 * |---1---|---1---|---1---|---1---|----3----|---3---|-----3-----|
 *  log_obj log_sub is_super can-   invul_int min_int  current_int
 *                           handle-
 *                           read error
 * 2. run_time_label:
 * |------3---------|----3---------|----3------------|------3--------|
 * | discr_invul_int| discr_min_int|discr_obj_min_int|current_min_int|
 *  
 * 3. new_cached_label:
 * |------3-------------|
 * | new_current_min_int|
 */
  

#define SRFD_XATTR_NAME "user.srfd" 

//#define SRFD_LABEL_LEN 6
#define SRFD_ZERO_INTEGRITY                         (0)
#define SRFD_DEFAULT_PHYSICAL_LABEL                 (7) 

#define SRFD_GET_CURRENT_INTEGRITY(label)           ((label.physical_label) & 0x0007)
#define SRFD_GET_MIN_INTEGRITY(label)               ((label.physical_label >> 3) & 0x0007)
#define SRFD_GET_INVUL_INTEGRITY(label)             ((label.physical_label >> 6) & 0x0007)
#define SRFD_GET_CAN_HANDLE_READ_ERROR(label)       ((label.physical_label >> 9) & 0x0001)
#define SRFD_GET_IS_SUPER(label)                    ((label.physical_label >> 9) & 0x0001)
#define SRFD_GET_LOG_SUB(label)                     ((label.physical_label >> 11) & 0x0001)  
#define SRFD_GET_LOG_OBJ(label)                     ((label.physical_label >> 12) & 0x0001)
#define SRFD_GET_CURRENT_MIN_INTEGRITY(label)       (label.run_time_label & 0x0007)
#define SRFD_GET_DISCR_OBJ_MIN_INTEGRITY(label)     ((label.run_time_label >> 3) & 0x0007)
#define SRFD_GET_DISCR_MIN_INTEGRITY(label)         ((label.run_time_label >> 6) & 0x0007)
#define SRFD_GET_DISCR_INVUL_INTEGRITY(label)       ((label.run_time_label >> 9) & 0x0007)
#define SRFD_GET_CACHED_CURRENT_MIN_INTEGRITY(label) (label.new_cached_label)
#define SRFD_GET_NEXT_INTEGRITY(a)                  (a+1)

#define SRFD_SET_CURRENT_INTEGRITY(label, new_value)       \
	(label.physical_label = ((label.physical_label & 0xFFF8) || (new_value & 0x0007)))
#define SRFD_SET_MIN_INTEGRITY(label, new_value)           \
	(label.physical_label = ((label.physical_label & 0xFFC7) || ((new_value & 0x0007) << 3)))
#define SRFD_SET_INVUL_INTEGRITY(label, new_value)         \
	(label.physical_label = ((label.physical_label & 0xFE3F) || ((new_value & 0x0007)<< 6)))
#define SRFD_SET_CAN_HANDLE_READ_ERROR(label, new_value)   \
	(label.physical_label = ((label.physical_label & 0xFDFF) || ((new_value & 0x0001) << 9)))
#define SRFD_SET_IS_SUPER(label, new_value)                \
	(label.physical_label = ((label.physical_label & 0xFBFF) || ((new_value & 0x0001) << 10)))
#define SRFD_SET_LOG_SUB(label, new_value)                 \
	(label.physical_label = ((label.physical_label & 0xF7FF) || ((new_value & 0x0001) << 11)))
#define SRFD_SET_LOG_OBJ(label, new_value)                 \
	(label.physical_label = ((label.physical_label & 0xEFFF) || ((new_value & 0x0001) << 12)))
#define SRFD_SET_CURRENT_MIN_INTEGRITY(label, new_value)   \
	(label.run_time_label = ((label.run_time_label & 0xFFF8) || (new_value & 0x0007)))
#define SRFD_SET_DISCR_OBJ_MIN_INTEGRITY(label, new_value) \
	(label.run_time_label = ((label.run_time_label & 0xFFC7) || ((new_value & 0x0007) << 3)))
#define SRFD_SET_DISCR_MIN_INTEGRITY(label, new_value)     \
	(label.run_time_label = ((label.run_time_label & 0xFE3F) || ((new_value & 0x0007) << 6)))
#define SRFD_SET_DISCR_INVUL_INTEGRITY(label, new_value)   \
	(label.run_time_label = ((label.run_time_label & 0xF1FF) || ((new_value & 0x0007) << 9)))
#define SRFD_SET_CACHED_CURRENT_MIN_INTEGRITY(label, new_value) \
    (label.new_cached_label = ((label.new_cached_label & 0xFFF8) || (new_value & 0x0007)))

#define SRFD_SET_PHYSICAL_LABEL(label, new_value)          \
	(label.physical_label = new_value)
#define SRFD_SET_RUN_TIME_LABEL(label, new_value)          \
    (label.run_time_label = new_value)
#define SRFD_SET_NEW_CACHED_LABEL(label, new_value)        \
    (label.new_cached_label = new_value)	



/* Following interface allows user to change the integrity levels.
 * Client only needs to re-compile the SRFD library
 */

#define SRFD_LABEL_MAX_INTEGRITY_LEVEL 7
#define SRFD_LABEL_MIN_INTEGRITY_LEVEL 0

#define MIN_INTEGRITY_LEVEL 0
#define SRFD_LABEL_LEN 8
#define MAX_LBL_BITS 3



#define SRFD_sub_group_lbl SRFD_subject_lbl



#define SRFD_label2int(_label) ((_label.level))


/* SRFD label abstract data structure */
struct SRFD_lbl {
	unsigned char level;
};

/* SRFD object labels */
struct SRFD_object_lbl {
	struct SRFD_lbl current_lbl;        /* object's current label */
	struct SRFD_lbl min_lbl;            /* local policy: current_lbl >= min_lbl */
	struct SRFD_lbl current_min_lbl;    /* derived policy: current_lbl >= current_min_lbl */ 
	struct SRFD_lbl new_current_min_lbl;        /* used internally for label propagation */
};

/* SRFD subject labels */
struct SRFD_subject_lbl {
	unsigned int can_handle_read_errors : 1;
	unsigned int can_handle_write_errors : 1;
	struct SRFD_lbl current_lbl;
	struct SRFD_lbl min_lbl;
	struct SRFD_lbl discr_min_lbl;
	struct SRFD_lbl new_current_min_lbl;


	struct SRFD_lbl discr_obj_min_lbl;
	struct SRFD_lbl invul_lbl;
	struct SRFD_lbl discr_invul_lbl;
	struct SRFD_lbl input_invul;
	struct SRFD_lbl discr_input_invul;
	struct SRFD_lbl output_invul;
	struct SRFD_lbl discr_output_invul;
	struct SRFD_lbl current_min_lbl;
	struct SRFD_lbl transient_current_lbl; /* for lookup */
	struct SRFD_lbl virtual_current_lbl; /* for symlinks */

	/* TODO: we might need new_discr_min_lbl */
};

/* SRFD file label */ 
struct SRFD_file_lbl {
	struct SRFD_object_lbl obj_lbl;
	struct SRFD_subject_lbl subj_lbl;
};

/* SRFD handle label */
struct SRFD_handle_lbl {
	struct SRFD_lbl current_lbl;
	struct SRFD_lbl current_min_lbl;
	struct SRFD_lbl invul_lbl;
	struct SRFD_lbl discr_min_lbl; /* For read handle only */
	struct SRFD_lbl new_current_min_lbl;
};


extern struct SRFD_object_lbl srfd_security_object_default_label;
extern struct SRFD_subject_lbl srfd_security_subject_default_label;



#if __KERNEL__
/* Operations on SRFD_lbl data structure*/

#define srfd_label_lub(l1, l2) ((l1.level > l2.level) ? l1 : l2)
#define srfd_label_glb(l1, l2) ((l1.level < l2.level) ? l1 : l2)
#define srfd_label_ge(l1, l2) (l1.level >= l2.level)
#define srfd_label_eq(l1, l2) (l1.level == l2.level)
#define srfd_label_gt(l1, l2) (l1.level > l2.level)
#define srfd_label_lt(l1, l2) (l1.level < l2.level)
#define srfd_label_le(l1, l2) (l1.level <= l2.level)

struct SRFD_lbl srfd_label_max_lbl(void);
struct SRFD_lbl srfd_label_min_lbl(void);



struct SRFD_lbl max_lbl(void);
struct SRFD_lbl min_lbl(void);
struct SRFD_lbl lub_lbl(struct SRFD_lbl, struct SRFD_lbl);
struct SRFD_lbl glb_lbl(struct SRFD_lbl, struct SRFD_lbl);
unsigned char geq_lbl(struct SRFD_lbl, struct SRFD_lbl);
unsigned char gt_lbl(struct SRFD_lbl, struct SRFD_lbl);
unsigned char eq_lbl(struct SRFD_lbl, struct SRFD_lbl);
char *serialize_lbl(struct SRFD_lbl);
struct SRFD_lbl new_lbl(char*); /* deserialize operation */

/* Operations on SRFD_object_lbl data structure */
char* serialize_object_lbl(struct SRFD_object_lbl);
struct SRFD_object_lbl new_object_lbl(char*);
void dump_object_label(struct SRFD_object_lbl *); /* debug helper function */

/* Operations on SRFD_subject_lbl data structure */
char* serialize_subject_lbl(struct SRFD_subject_lbl);
void dump_subject_label(struct SRFD_subject_lbl *); /* debug helper function */
struct SRFD_subject_lbl new_subject_lbl(char*); /* deserialize operation */

/* Operations on SRFD_file_lbl data structure */
char* serialize_file_lbl(struct SRFD_file_lbl);
struct SRFD_file_lbl new_file_lbl(const char*); /* deserialize operation */


/* Dump Handle Label */
void dump_handle_label(struct SRFD_handle_lbl * lbl);
#endif

//#define SRFD_LABEL_BASEDONNEW 1
//#define SRFD_LABEL_BASEDONCURRENT 0

#endif /* __SRFD_LABEL_H__ */

