/* * Self-Revocation Free Downgrading - A kernel module that applies information flow
 * control to protect system integrity without self-revocation
 * Copyright (C) 2014 Secure Systems Laboratory, Stony Brook University
 *
 * This file is part of Self-Revocation Free Downgrading.
 *
 * Self-Revocation Free Downgrading is free software: you can redistribute it
 * and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Self-Revocation Free Downgrading is distributed in the hope that it will be
 * useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Self-Revocation Free Downgrading.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

#define _GNU_SOURCE
#include <dlfcn.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>

#include <stdlib.h>
#include <string.h>
#include <limits.h>

#include <sys/sendfile.h>
#include <errno.h>

#include <unistd.h> 

#include <stdarg.h>

#include "ppiu_debug.h"


/* If the original action failed, then check if it involves write. If the path involved is 
	redirectable, change it to the shadow directory and retry
*/

#define declareCallType(_origName) typedef _origName ##_retType (*orig_ ##_origName ##_type) _origName ##_param

#define open_param (const char *pathname, int flags, ...)
#define open_retType int
declareCallType(open);

#define fopen_param (const char *path, const char *mode)
#define fopen_retType FILE *
declareCallType(fopen);

#define open64_param (const char *pathname, int flags, ...)
#define open64_retType int
declareCallType(open64);

#define rename_param (const char *oldpath, const char *newpath)
#define rename_retType int
declareCallType(rename);


#define declareCall(_origName) _origName ##_retType _origName _origName ##_param


#define getOriginalCall(_origName, _origFunc) orig_ ##_origName ##_type _origFunc; \
	_origFunc = (orig_ ##_origName ##_type) dlsym(RTLD_NEXT, #_origName)


#define REDIRECT_DIR "/tmp/redirect"


__thread char *privateRedirectPath = NULL;


const char * const ppi_redirect_redirectableDirs[] = {
	"/run/user/1000/dconf/user",
	"/home/ubuntu/",
	NULL
};



int ppi_redirect_isRedirectable(const char *path) {
	int count = 0;

	while (ppi_redirect_redirectableDirs[count] != NULL) {
		if (strncmp(path, ppi_redirect_redirectableDirs[count], strlen(ppi_redirect_redirectableDirs[count])) == 0)
			return 1;
		count++;
	}
	return 0;
}




char *getRedirectedPath(const char *path) {
	//Cannot allocate and deallocate. gnome-calculator failed!!!
	if (privateRedirectPath == NULL)
		privateRedirectPath = malloc(PATH_MAX);

	sprintf(privateRedirectPath, REDIRECT_DIR "%s", path);

	return privateRedirectPath;

}



int ppi_redirect_shadowCopyExists(const char *path) {
	char *redirectedPath = getRedirectedPath(path);
	return (access(redirectedPath, F_OK) == 0);
}



int ppi_utils_createDirectories(const char *path) {
	char *tempPath = NULL;
	int count = 0;

	if (path == NULL) return -1;

	tempPath = malloc(strlen(path) + 1);
	while (path[count] != '\0') {
		tempPath[count] = path[count];
		if (tempPath[count] == '/' && count) {
			tempPath[count + 1] = '\0';
			if (mkdir(tempPath, S_IRWXU|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH) && errno != EEXIST)
				PPIU_LOG(ERROR, "Create directory error %d", errno);
		}
		count++;
	}
	if (tempPath) free(tempPath);
	return 0;
}


int ppi_utils_copyFileFD(int srcFD, int dstFD) {
	struct stat buf;
	off_t fileSize, offset = 0, totalByteCopied = 0, byteCopied;

	if (fstat(srcFD, &buf)) return -1;
	fileSize = buf.st_size;

	while (totalByteCopied < fileSize) {
		if ((byteCopied = sendfile(dstFD, srcFD, &offset, fileSize - totalByteCopied)) <= 0) {
			if (errno == EINTR || errno == EAGAIN)
				continue;
			PPIU_LOG(ERROR, "Failed to copy file using sendfile: errno %d", errno);
			return -1;
		}
		totalByteCopied += byteCopied;
	}
	return 0;
}


/* When the src does not exists, then no need to copy. Just create a shadow copy 
    e.g., when in .recently-used.xxxxxx
*/
int ppi_utils_createShadowCopy(const char *path) {
	int srcFD, dstFD;
	int rv = -1;

	getOriginalCall(open, orig_open);
	char *redirectedPath = getRedirectedPath(path);

	srcFD = orig_open(path, O_RDONLY);
	if (srcFD == -1) goto out;

	ppi_utils_createDirectories(redirectedPath);
	dstFD = orig_open(redirectedPath, O_WRONLY|O_CREAT, S_IRWXU|S_IRGRP|S_IXGRP);
	PPIU_LOG(INFO, "open destination file %s gave %d", redirectedPath, dstFD);

	if (dstFD == -1) goto closeSrcFD;

	rv = ppi_utils_copyFileFD(srcFD, dstFD);
	if (rv)
		PPIU_LOG(ERROR, "Failed to copy file %s", path);

	close(dstFD);
closeSrcFD:
	if (srcFD != -1) close(srcFD);
out:
	return rv;
}


#define open_flagsHasWrite(_flags) ((_flags & O_WRONLY) == O_WRONLY || (_flags & O_RDWR) == O_RDWR)

declareCall(open) {

	int rv;
	getOriginalCall(open, orig_open);
	va_list args;
	va_start(args, flags);
	mode_t mode = va_arg(args, mode_t);

	rv = orig_open(pathname, flags, mode);

	PPIU_LOG(INFO, "[open] Opening file %s gives %d", pathname, rv);

	if (rv == -1 && (errno == EACCES || errno == EPERM) && open_flagsHasWrite(flags) && ppi_redirect_isRedirectable(pathname)) {
		PPIU_LOG(INFO, "Creating shadow copy");
		if (ppi_utils_createShadowCopy(pathname) == 0) {
			char *redirectedPath;
//openShadowCopy:
			redirectedPath = getRedirectedPath(pathname);
			rv = orig_open(redirectedPath, flags, mode);
			PPIU_LOG(INFO, "Redirected open %s gives %d, errno %d", redirectedPath, rv, errno);
		}
	}

	va_end(args);

	return rv;
		
}

declareCall(open64) {
	int rv;
	getOriginalCall(open64, orig_open64);

	va_list args;
	va_start(args, flags);
	mode_t mode = va_arg(args, mode_t);

	rv = orig_open64(pathname, flags, mode);

	PPIU_LOG(INFO, "[open64] Opening file %s gives %d", pathname, rv);

	if (rv == -1 && (errno == EACCES || errno == EPERM) && open_flagsHasWrite(flags) && ppi_redirect_isRedirectable(pathname)) {
		if (flags & O_CREAT) {
			/* If the program wants to create the file, change it to the redirected path
			   without making a shadow copy 
			 */
			char *redirectedPath;
			redirectedPath = getRedirectedPath(pathname);
			ppi_utils_createDirectories(redirectedPath);
			rv = orig_open64(redirectedPath, flags, args);
		} else {
			PPIU_LOG(INFO, "Creating shadow copy");
			if (ppi_utils_createShadowCopy(pathname) == 0) {
				char *redirectedPath;
//openShadowCopy:
				redirectedPath = getRedirectedPath(pathname);
				rv = orig_open64(redirectedPath, flags, mode);

				PPIU_LOG(INFO, "Redirected open gives %d", rv);
			}
		}
	}

	va_end(args);

	return rv;
}




int fopen_modeHasWrite(const char *mode) {
	int count = 0;

	if (mode == NULL) return 0;
	while (mode[count] != '\0') {
		switch (mode[count]) {
		case 'w':
		case '+':
		case 'a':
			return 1;
		}
		count++;
	}
	return 0;
}



declareCall(fopen) {
	FILE *file;

	getOriginalCall(fopen, orig_fopen);

	file = orig_fopen(path, mode);
	PPIU_LOG(INFO, "[fopen] Opening file %s gives %p", path, file);

	if (file == NULL && (errno == EACCES || errno == EPERM) && fopen_modeHasWrite(mode) && ppi_redirect_isRedirectable(path)) {
		PPIU_LOG(INFO, "Creating shadow copy");
		if (ppi_utils_createShadowCopy(path) == 0) {
			char *redirectedPath;
			redirectedPath = getRedirectedPath(path);
			file = orig_fopen(redirectedPath, mode);
			PPIU_LOG(INFO, "Redirected open gives %p", file);
		}
	}

	return file;
}




declareCall(rename) {
	int rv;
	getOriginalCall(rename, orig_rename);

	const char *modifiedOldPath = oldpath;
	char *bufferModifiedOldPath = NULL;
	const char *modifiedNewPath = newpath; //, *bufferModifiedNewPath = NULL;

	if (ppi_redirect_shadowCopyExists(oldpath)) {
		bufferModifiedOldPath = malloc(PATH_MAX);
		strncpy(bufferModifiedOldPath, getRedirectedPath(oldpath), PATH_MAX);
		modifiedOldPath = bufferModifiedOldPath;
	}

	rv = orig_rename(modifiedOldPath, modifiedNewPath);

	if (rv == -1 && (errno == EACCES || errno == EPERM) && ppi_redirect_isRedirectable(modifiedNewPath)) {
		/* If fail, may be because the new path is not writable */
		modifiedNewPath = getRedirectedPath(newpath);
		ppi_utils_createDirectories(modifiedNewPath);
		rv = orig_rename(modifiedOldPath, modifiedNewPath);
	}

	PPIU_LOG(INFO, "rename is called from %s to %s, returned %d", modifiedOldPath, modifiedNewPath, rv);

	if (bufferModifiedOldPath != NULL) free(bufferModifiedOldPath);

	return rv;
}



