/* * Self-Revocation Free Downgrading - A kernel module that applies information flow
 * control to protect system integrity without self-revocation
 * Copyright (C) 2014 Secure Systems Laboratory, Stony Brook University
 *
 * This file is part of Self-Revocation Free Downgrading.
 *
 * Self-Revocation Free Downgrading is free software: you can redistribute it
 * and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Self-Revocation Free Downgrading is distributed in the hope that it will be
 * useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Self-Revocation Free Downgrading.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

#include "srfd_abstract.h"
#include "srfd_security.h"
#include "srfd_security_internal.h"
#include "srfd_debug.h"
#include "srfd_internal.h"
#include <linux/mm.h>

struct SRFD_handle *srfd_openNewHandle(struct SRFD_subgroup *subgroup, struct SRFD_object *object,
	SRFD_HANDLE_TYPE handleType, int flow_mediated, int delayed) {

	struct SRFD_lbl handle_invul_lbl, subgroup_current_lbl, object_current_lbl,
			handle_current_lbl, upstream_current_min_lbl;
	struct SRFD_handle *newHandle = NULL;


	srfd_assert1(subgroup != NULL);
	srfd_assert1(object != NULL);

	srfd_info("Subgroup %p current_lbl: %d, current_min_lbl is %d, new min: %d", subgroup,
		srfd_get_subgroupLabelV(subgroup, current_lbl), srfd_get_subgroupLabelV(subgroup, current_min_lbl),
		srfd_get_subgroupLabelV(subgroup, new_current_min_lbl));

	subgroup_current_lbl = srfd_get_subgroupLabel(subgroup, current_lbl);
	object_current_lbl = srfd_get_objectLabel(object, current_lbl);

	newHandle = srfd_get_handle(srfd_handle_alloc());
	if (srfd_assert1r(newHandle != NULL))
		goto out;

	newHandle->h_type = handleType;
	newHandle->h_isDelayed = delayed;
	newHandle->h_isFlowMediated = flow_mediated;
	newHandle->h_fd = -1;
	newHandle->h_o = object;	/* need to fill in the details first for propagation*/
	newHandle->h_sg = subgroup;

	switch(handleType) {
	case SRFD_HANDLE_TYPE_READ:
		newHandle->h_canHandleErrors = (srfd_subgroup_canHandleReadError(subgroup) &&
				flow_mediated) ? 1 : 0;
		srfd_set_handleLabel(newHandle, discr_min_lbl,
				srfd_get_subgroupLabel(subgroup, discr_min_lbl));
		handle_invul_lbl = srfd_get_subgroupLabel(subgroup, discr_input_invul);

		/* If the handle's input invul_lbl is less than subject's current_lbl,
		 * then the label can remain to be subject's current_lbl as long as
		 * the object's label is not below the handle's invul_lbl */
		if (srfd_label_lt(handle_invul_lbl, subgroup_current_lbl) &&
				(srfd_label_ge(handle_invul_lbl, object_current_lbl))) {
			handle_current_lbl = subgroup_current_lbl;
			srfd_info("handle is read, setting handle lbl as subgroup lbl: %d", subgroup_current_lbl.level);
		} else {
			handle_current_lbl = object_current_lbl;
			srfd_info("handle is read, setting handle lbl as object lbl: %d", object_current_lbl.level);
		}
		/* min lbl constraint is propogated from subgroup if the object is opened for reading */
		upstream_current_min_lbl = srfd_get_subgroupLabel(subgroup, current_min_lbl);
		break;
	case SRFD_HANDLE_TYPE_WRITE:
		newHandle->h_canHandleErrors = (srfd_subgroup_canHandleWriteError(subgroup) && 
				flow_mediated) ? 1 : 0;
		handle_invul_lbl = srfd_get_subgroupLabel(subgroup, discr_output_invul);

		/* If subject current_lbl is at least the output invul_lbl,
		 * handle will have object's label, otherwise, handle will
		 * have subject's current_lbl */
		if (srfd_label_ge(subgroup_current_lbl, handle_invul_lbl)) {
			handle_current_lbl = object_current_lbl;
			srfd_info("handle is write, setting handle lbl as object lbl: %d", object_current_lbl.level);
		} else {
			handle_current_lbl = subgroup_current_lbl;
			srfd_info("handle is write, setting handle lbl as subgroup lbl: %d", subgroup_current_lbl.level);
		}

		upstream_current_min_lbl = srfd_get_objectLabel(object, current_min_lbl);
		break;
	default:
		srfd_error("Unknown handle type");
		goto close_handle;
	}

	srfd_set_handleLabel(newHandle, invul_lbl, handle_invul_lbl);
	srfd_set_handleLabel(newHandle, current_lbl, handle_current_lbl);
	srfd_set_handleLabel(newHandle, current_min_lbl, upstream_current_min_lbl);
	srfd_set_handleLabel(newHandle, new_current_min_lbl, upstream_current_min_lbl);

	srfd_info("handle_current_lbl is %d", srfd_get_handleLabelV(newHandle, current_lbl));
	srfd_info("Subgroup %p current_min_lbl is %d, new min: %d", subgroup, srfd_get_subgroupLabelV(subgroup, current_min_lbl), srfd_get_subgroupLabelV(subgroup, new_current_min_lbl));
	srfd_info("upstream_current_min_lbl is %d", upstream_current_min_lbl.level);



	switch(handleType) {
	case SRFD_HANDLE_TYPE_READ:
		/* Should also consider increasing current_min_lbl of object if cannot decrease subgroup !!! */
		if (srfd_current_min_lbl_subgroupCanDecrease2Target(subgroup, srfd_get_objectLabel(object, current_min_lbl)) == -1) {
			srfd_logViolation("Opening of READ object ino %lu for subgroup %p is prevented by min label", object->o_ino, subgroup);

			if (srfd_current_min_lbl_objectCanIncrease2Target(object, srfd_get_subgroupLabel(subgroup, current_min_lbl)) == 1) {
				srfd_logViolation("Opening of READ object ino %lu for subgroup %p can be allowed by increasing object current_min_lbl", object->o_ino, subgroup);
				srfd_current_min_lbl_updateWithNewFromObject(object);
				srfd_info("Opening is allowed");
				break;
			} else {
				srfd_logTEMP("object current_min_lbl cannot be increased to solve the constraint");
			}


#ifdef SRFD_VERBOSE_LOGGING
			srfd_printSubgroupInfo2(srfd_logViolation, subgroup);
			srfd_printSubgroupWriteObjects(srfd_logViolation, subgroup);
			srfd_printObjectInfo2(srfd_logViolation, object);
#endif
			goto close_handle;
		} else {
			srfd_current_min_lbl_updateWithNewFromSubgroup(subgroup);
			srfd_info("Subgroup %p Opening for read is allowed", subgroup);
#ifdef SRFD_VERBOSE_LOGGING
			srfd_printSubgroupInfo2(srfd_logViolation, subgroup);
			srfd_printSubgroupWriteObjects(srfd_info, subgroup);
			srfd_printObjectInfo2(srfd_logViolation, object);
#endif
		}
		break;

	case SRFD_HANDLE_TYPE_WRITE:
		if (srfd_current_min_lbl_subgroupCanIncrease2Target(subgroup, srfd_get_objectLabel(object, current_min_lbl)) == -1) {
			srfd_logViolation("Opening of WRITE object ino %lu for subgroup %p is prevented by min label", object->o_ino, subgroup);

			if (srfd_current_min_lbl_objectCanDecrease2Target(object, srfd_get_subgroupLabel(subgroup, current_min_lbl)) == 1) {
				srfd_logViolation("Opening of WRITE object ino %lu for subgroup %p can be allowed by decreasing object current_min_lbl", object->o_ino, subgroup);
				srfd_current_min_lbl_updateWithNewFromObject(object);
				srfd_info("Opening is allowed");
				break;
			} else {
				srfd_logTEMP("object current_min_lbl cannot be decreased to solve the constraint");
			}

#ifdef SRFD_VERBOSE_LOGGING
			srfd_printSubgroupInfo2(srfd_logViolation, subgroup);
			srfd_printSubgroupWriteObjects(srfd_logViolation, subgroup);
			srfd_printObjectInfo2(srfd_logViolation, object);
#endif
			goto close_handle;
		} else {
			srfd_current_min_lbl_updateWithNewFromSubgroup(subgroup);
			srfd_info("Opening for writing is allowed");

		}	
		break;
	default:
		srfd_error("Unhandled handle type");
	}
	srfd_addHandle2Object(newHandle, object);
	srfd_addHandle2Subgroup(newHandle, subgroup);

out:
	return newHandle;



close_handle:
	srfd_put_handle(newHandle);
	newHandle = NULL;
	goto out;
}




int srfd_abstract_open(struct SRFD_subgroup *subgroup, struct SRFD_object *object, int mask) {

        unsigned int flow_mediated, delayed;
	struct SRFD_handle *newReadHandle = NULL;
        struct SRFD_handle *newWriteHandle = NULL;

	struct SRFD_handle *handle;

	int ret = 0;


	if (down_interruptible(&srfd_open_sem))
		srfd_info("++++++ Failed to grab semaphore");

	preempt_disable();


#if SRFD_POLICY == SRFD_POLICY_NODOWNGRADING
	if (mask & MAY_READ)
		ret = srfd_label_lt(srfd_get_objectLabel(object, current_lbl),
				srfd_get_subgroupLabel(subgroup, current_lbl));
	if (ret) {
		srfd_logViolation("open for reading is denied");
		ret = -EACCES;
		goto out;
	}
	if (mask & MAY_WRITE)
		ret = srfd_label_lt(srfd_get_subgroupLabel(subgroup, current_lbl),
				srfd_get_objectLabel(object, current_lbl));
	if (ret) {
		srfd_logViolation("open for writing is denied");
		ret = -EACCES;
	}
	goto out;
#endif


        switch((object->o_inode->i_mode) & S_IFMT) {
                case S_IFLNK:
                case S_IFREG:
                case S_IFDIR:
                case S_IFCHR:
                case S_IFBLK:
                case S_IFIFO:
                                flow_mediated = 1;
                                delayed = 1;
                                break;
                default:/* applies to S_IFSOCK */
                                flow_mediated = 0;
                                delayed = 0;
        }

	/* If the file is already opened, do nothing.
	   Otherwise, call srfd_openNewHandle */
	if (mask & MAY_READ) {	
		srfd_do_foreach_subgroup_valid_rh_rm_invalid(subgroup, handle) {
			if (handle->h_o->o_inode == object->o_inode) {
				srfd_beforebreak_foreach_subgroup_rh(subgroup);
				goto openWrite;
			}
		} srfd_while_foreach_subgroup_rh(subgroup);

		newReadHandle = srfd_openNewHandle(subgroup, object, SRFD_HANDLE_TYPE_READ,
			flow_mediated, delayed);
		/* If newReadHandle == NULL when min label is not used, it is likely due
		   to errors other than permissions */
		if (newReadHandle == NULL) {
			ret = -EACCES;
			srfd_alwayslog("Opening for read ino %lu is not allowed", object->o_inode->i_ino);
			goto out;
		}
		srfd_info("subgroup %p open ino %lu for reading is allowed", subgroup, object->o_inode->i_ino);
	}

openWrite:

	if (mask & MAY_WRITE) {
		srfd_do_foreach_subgroup_valid_wh_rm_invalid(subgroup, handle) {
		if (handle->h_o->o_inode == object->o_inode) {
				srfd_beforebreak_foreach_subgroup_wh(subgroup);
				goto out;
			}
		} srfd_while_foreach_subgroup_wh(subgroup);

		newWriteHandle = srfd_openNewHandle(subgroup, object, SRFD_HANDLE_TYPE_WRITE,
			flow_mediated, delayed);

		if (newWriteHandle == NULL) {
			srfd_alwayslog("Opening for write ino %lu is not allowed", object->o_inode->i_ino);
			ret = -EACCES;
			if (newReadHandle != NULL) {
				srfd_info("Also closing the created read handle");
				srfd_removeHandleFromSubgroup(newReadHandle);
				srfd_removeHandleFromObject(newReadHandle);
				srfd_put_handle(newReadHandle);
			}
			goto out;
		}
		srfd_info("subgroup %p open ino %lu for writing is allowed", subgroup, object->o_inode->i_ino);
	}

out:

	preempt_enable();
	up(&srfd_open_sem);

	return ret;
}


int srfd_abstract_read(struct SRFD_handle *handle) {

	struct SRFD_lbl handle_current_lbl;
	struct SRFD_subgroup *subgroup;
	int ret = 0;
#ifdef SRFD_VERBOSE_LOGGING
	struct SRFD_object *object;
#endif
	handle_current_lbl = srfd_get_handleLabel(handle, current_lbl);

	subgroup = srfd_get_subgroup(handle->h_sg);
	if (srfd_assert1r(subgroup != NULL))
		goto out;

	srfd_info("checking if downgrade is needed");

#ifdef SRFD_POLICY_ENFORCE_MIN_LABEL
	/* If not using min label, read is always allowed */
	if (srfd_label_ge(handle_current_lbl, srfd_get_subgroupLabel(subgroup, current_min_lbl))) {
#endif

		if (srfd_label_lt(handle_current_lbl, srfd_get_subgroupLabel(subgroup, current_lbl))) {
			if (SRFD_IS_INTERESTING_PROC)
				srfd_alwayslog("need to downgrade subgroup");

			srfd_logDowngrade("Downgrading subgroup %p on reading object ino %lu:", subgroup, handle->h_o->o_ino);
			srfd_printSubgroupInfo2(srfd_logDowngrade, subgroup);
			srfd_printObjectInfo2(srfd_logDowngrade, handle->h_o);

#ifdef SRFD_VERBOSE_LOGGING

			srfd_printSubgroupWriteObjects(srfd_logDowngrade, subgroup);

			object = srfd_get_object(handle->h_o);
			srfd_assert1(object != NULL);
			if (object) {
				srfd_printObjectInfo2(srfd_logDowngrade, object);
				srfd_put_object(object);
			}

#endif

			srfd_set_subgroupLabel(subgroup, current_lbl, handle_current_lbl);
			srfd_current_lbl_propagateFromSubgroup(subgroup);

		} else {
			srfd_info("no downgrade is needed: handle: %d, subgroup: %d",
				handle_current_lbl.level, srfd_get_subgroupLabelV(subgroup, current_lbl));
		}
#ifdef SRFD_POLICY_ENFORCE_MIN_LABEL
	} else {
		srfd_alwayslog("handle label is less than subgroup's cur min lbl");
		ret = -EACCES;
	}
#endif
	srfd_put_subgroup(subgroup);
out:
	return ret;

}

struct SRFD_object_lbl srfd_object_defaultLabel2 = {
	.current_lbl = {7},
	.min_lbl = {1}
};




int srfd_abstract_write(struct SRFD_handle *handle) {

	struct SRFD_lbl handle_current_lbl;
	struct SRFD_object *object;

	int ret = 0;

	handle_current_lbl = srfd_get_handleLabel(handle, current_lbl);

	object = srfd_get_object(handle->h_o);
	if (srfd_assert1r(object != NULL))
		goto out;

	if (srfd_label_ge(handle_current_lbl, srfd_get_objectLabel(object, current_min_lbl))) {
		/* Downgrade object */
		if (srfd_label_lt(handle_current_lbl, srfd_get_objectLabel(object, current_lbl))) {

#ifdef SRFD_VERBOSE_LOGGING
			struct SRFD_subgroup *subgroup;

			srfd_logDowngrade("Object ino %lu has to be downgraded due to write from subgroup %p", object->o_ino, handle->h_sg);

			srfd_printObjectInfo2(srfd_logDowngrade, object);
			subgroup = srfd_get_subgroup(handle->h_sg);
			srfd_assert1(subgroup != NULL);
			if (subgroup != NULL) {
				srfd_printSubgroupInfo2(srfd_logDowngrade, subgroup);
				srfd_put_subgroup(subgroup);
			}
#endif
			srfd_assert(srfd_label_le(srfd_get_objectLabel(object, min_lbl),
				handle_current_lbl), "min lbl: %d, current lbl: %d", srfd_get_objectLabelV(object, min_lbl), SRFD_label2int(handle_current_lbl));

			srfd_set_objectLabel(object, current_lbl, handle_current_lbl);
			srfd_objectDowngraded(object, handle_current_lbl);
			srfd_current_lbl_propagateFromObject(object);
		}
	} else {
		srfd_alwayslog("handle label is less than object's cur min lbl");
		/* LOMAC will fail here, and object cannot be downgraded */
#ifndef SRFD_POLICY_ENFORCE_MIN_LABEL
		srfd_logViolation("LOMAC prevents write to object!!!");
#else
		srfd_error("Should not be reachable");
#endif
		ret = -EACCES;
	}

	srfd_put_object(object);
out:
	return ret;

}


int srfd_abstract_unlink(struct SRFD_subgroup *subgroup, struct SRFD_object *object) {

	int ret = 0;

	if (srfd_label_lt(srfd_get_subgroupLabel(subgroup, current_lbl),
		srfd_get_objectLabel(object, current_lbl))) {
		srfd_logViolation("Unlink: subgroup %p, level %d, ino: %lu, level %d",
			subgroup, srfd_get_subgroupLabelV(subgroup, current_lbl),
			object->o_ino, srfd_get_objectLabelV(object, current_lbl));
		ret = -EACCES;
	}
	return ret;
}


/* Modifications to directory may involve only a single syscall, e.g., unlink.
   Hence, no valid handle will be maintained. Need to perform the action in one shot. */

int srfd_abstract_dirWrite(struct SRFD_subgroup *subgroup, struct SRFD_object *dirObject) {

	int rv = srfd_abstract_open(subgroup, dirObject, MAY_READ|MAY_WRITE);
	if (rv == 0) {
		srfd_assert1(srfd_label_ge(srfd_get_objectLabel(dirObject, min_lbl),
					srfd_get_subgroupLabel(subgroup, current_lbl)));
		srfd_logDowngrade("subgroup lbl: %d, dir lbl: %d", srfd_get_subgroupLabelV(subgroup, current_lbl), srfd_get_objectLabelV(dirObject, current_lbl));
		if (srfd_label_ge(srfd_get_objectLabel(dirObject, current_lbl), 
					srfd_get_subgroupLabel(subgroup, current_lbl))) {
			srfd_set_objectLabel(dirObject, current_lbl, srfd_get_subgroupLabel(subgroup, current_lbl));
			srfd_current_lbl_propagateFromObject(dirObject);
		} else {
			rv = -EPERM;
			srfd_logViolation("Deny subgroup %p from modifying directory of ino %lu", subgroup, dirObject->o_ino);
			srfd_printSubgroupInfo2(srfd_logViolation, subgroup);
			srfd_printObjectInfo2(srfd_logViolation, dirObject);
		}
	}
	return rv;
}




