/* * Self-Revocation Free Downgrading - A kernel module that applies information flow
 * control to protect system integrity without self-revocation
 * Copyright (C) 2014 Secure Systems Laboratory, Stony Brook University
 *
 * This file is part of Self-Revocation Free Downgrading.
 *
 * Self-Revocation Free Downgrading is free software: you can redistribute it
 * and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Self-Revocation Free Downgrading is distributed in the hope that it will be
 * useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Self-Revocation Free Downgrading.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

#include "srfd_security.h"
#include "srfd_security_internal.h"


static DEFINE_SPINLOCK(srfd_credsec_list_lock);

struct list_head srfd_credsec_list[SRFD_MAX_CREDSEC_HASH];
atomic_t srfd_credsec_activeCount = ATOMIC_INIT(0);

#define cred2hash(credPtr) \
	&srfd_credsec_list[hash_ptr((void *)credPtr, SRFD_CREDSEC_HASH_BITS)]


void __srfd_credsec_add2List(struct SRFD_credsec *ptr) {
	unsigned long flags;
	struct list_head *list = cred2hash(ptr->cs_cred);

	spin_lock_irqsave(&srfd_credsec_list_lock, flags);
	list_add_rcu(&ptr->cs_list, list);
	spin_unlock_irqrestore(&srfd_credsec_list_lock, flags);
	atomic_inc(&srfd_credsec_activeCount);
}

static void __srfd_credsec_rcuFree(struct rcu_head *rcu) {
	struct SRFD_credsec *ptr = container_of(rcu, typeof(*ptr), cs_rcu);
	kfree(ptr);
	atomic_dec(&srfd_credsec_activeCount);
}

void srfd_credsec_free(struct SRFD_credsec *credsec) {
	
	unsigned long flags;

	if (atomic_read(&credsec->cs_rc) > 1)
		srfd_info("Freeing credsec with usage count: %d > 1 %p", atomic_read(&credsec->cs_rc), credsec);

	if (credsec->cs_parent_cs != NULL) {
		srfd_put_credsec(credsec->cs_parent_cs);
		credsec->cs_parent_cs = NULL;
	}

	spin_lock_irqsave(&srfd_credsec_list_lock, flags);
	list_del_rcu(&credsec->cs_list);
	spin_unlock_irqrestore(&srfd_credsec_list_lock, flags);
	call_rcu(&credsec->cs_rcu, __srfd_credsec_rcuFree);
}

struct SRFD_credsec *srfd_get_credsec(struct SRFD_credsec *credsec) {
	if (credsec != NULL)
		atomic_inc(&credsec->cs_rc);
	return credsec;
}

void srfd_put_credsec(struct SRFD_credsec *credsec) {
	if (atomic_dec_and_test(&credsec->cs_rc))
		srfd_credsec_free(credsec);
}

struct SRFD_credsec *srfd_credsec_findByCred(const struct cred *p) {
	struct list_head *list = cred2hash(p);
	struct SRFD_credsec *credsec;

	while (!list->next);
	rcu_read_lock();
	list_for_each_entry_rcu(credsec, list, cs_list) {
		if (credsec->cs_cred != p)
			continue;
		rcu_read_unlock();
		return credsec;
	}

	rcu_read_unlock();
	return NULL;
}

struct SRFD_credsec *srfd_credsec_allocAndGet(const struct cred *p) {

	struct SRFD_credsec *credsec = kzalloc(sizeof(*credsec), GFP_ATOMIC);

	if (srfd_assert1r(credsec != NULL))
		return NULL;

	atomic_set(&credsec->cs_rc, 1);
	credsec->cs_cred = p;
	credsec->cs_parent_cs = NULL;

	__srfd_credsec_add2List(credsec);

	return credsec;
}

