/* * Self-Revocation Free Downgrading - A kernel module that applies information flow
 * control to protect system integrity without self-revocation
 * Copyright (C) 2014 Secure Systems Laboratory, Stony Brook University
 *
 * This file is part of Self-Revocation Free Downgrading.
 *
 * Self-Revocation Free Downgrading is free software: you can redistribute it
 * and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Self-Revocation Free Downgrading is distributed in the hope that it will be
 * useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Self-Revocation Free Downgrading.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

#include "srfd_security.h"
#include "srfd_debug.h"

struct SRFD_lbl srfd_computeHandle_current_lbl(struct SRFD_handle *handle) {

	struct SRFD_lbl handle_current_lbl, subgroup_current_lbl, object_current_lbl, handle_invul_lbl;
	struct SRFD_subgroup *subgroup;
	struct SRFD_object *object;

	handle_current_lbl = srfd_label_max_lbl();

	if (srfd_get_handle(handle) == NULL) {
		srfd_error("handle is null");
		goto out;
	}
	handle_invul_lbl = srfd_get_handleLabel(handle, invul_lbl);

	subgroup = srfd_get_subgroup(handle->h_sg);
	if (subgroup == NULL) {
		srfd_error("subgroup is null");
		goto put_handle;
	}
	subgroup_current_lbl = srfd_get_subgroupLabel(subgroup, current_lbl);

	object = srfd_get_object(handle->h_o);
	if (object == NULL) {
		srfd_error("object is NULL");
		goto put_subgroup;
	}
	object_current_lbl = srfd_get_objectLabel(object, current_lbl);

	switch(handle->h_type) {
	case SRFD_HANDLE_TYPE_READ:
		if (srfd_label_lt(handle_invul_lbl, subgroup_current_lbl) &&
				(srfd_label_ge(handle_invul_lbl, object_current_lbl)))
			handle_current_lbl = subgroup_current_lbl;
		else
			handle_current_lbl = object_current_lbl;
		break;
	case SRFD_HANDLE_TYPE_WRITE:
		if (srfd_label_ge(subgroup_current_lbl, handle_invul_lbl))
			handle_current_lbl = object_current_lbl;
		else
			handle_current_lbl = subgroup_current_lbl;
		break;
	default:
		srfd_error("Unknown handle type");

	}

//put_object:
	srfd_put_object(object);
put_subgroup:
	srfd_put_subgroup(subgroup);
put_handle:
	srfd_put_handle(handle);
out:
	return handle_current_lbl;

}



void srfd_current_lbl_propagateFromSubgroup(struct SRFD_subgroup *subgroup) {

	struct SRFD_handle *wh;
	struct SRFD_lbl new_current_lbl;
	struct SRFD_lbl old_current_lbl;

	if (srfd_assert1r(srfd_get_subgroup(subgroup) != NULL))
		return;

	srfd_do_foreach_subgroup_valid_wh(subgroup, wh) {
		srfd_get_handle(wh);
		old_current_lbl = srfd_get_handleLabel(wh, current_lbl);

		if (wh->h_sg != subgroup)
			srfd_alwayslog("subgroup is %p, wh->subgroup is %p", subgroup, wh->h_sg);

		new_current_lbl = srfd_computeHandle_current_lbl(wh);
		if (srfd_label_gt(old_current_lbl, new_current_lbl)) {
			srfd_set_handleLabel(wh, current_lbl, new_current_lbl);
			//If flow is not mediated, need to propogate now, otherwise, wait until write
			if (!wh->h_isFlowMediated) {
				srfd_current_lbl_propagateFromHandle(wh);
			}
		}
		srfd_put_handle(wh);
	} srfd_while_foreach_subgroup_wh(subgroup);

	srfd_put_subgroup(subgroup);

}


void srfd_current_lbl_propagateFromObject(struct SRFD_object *object) {

	struct SRFD_handle *rh;
	struct SRFD_lbl new_current_lbl;
	struct SRFD_lbl old_current_lbl;

	srfd_do_foreach_object_valid_rh(object, rh) {

		srfd_get_handle(rh);

		if (rh->h_sg == NULL) {
			srfd_put_handle(rh);
			continue;
		}

		old_current_lbl = srfd_get_handleLabel(rh, current_lbl);
		new_current_lbl = srfd_computeHandle_current_lbl(rh);
		if (srfd_label_gt(old_current_lbl, new_current_lbl)) {
			srfd_set_handleLabel(rh, current_lbl, new_current_lbl);
			if (!rh->h_isFlowMediated) {
				srfd_current_lbl_propagateFromHandle(rh);
			}
		}
		srfd_put_handle(rh);
	} srfd_while_foreach_object_rh(object);
}


void srfd_current_lbl_propagateFromHandle(struct SRFD_handle *handle) {

	struct SRFD_lbl handle_current_lbl;
	struct SRFD_subgroup *subgroup;
	struct SRFD_object *object;

	if (srfd_get_handle(handle) == NULL)
		goto out;

	handle_current_lbl = srfd_get_handleLabel(handle, current_lbl);

	subgroup = srfd_get_subgroup(handle->h_sg);
	if (subgroup == NULL) {
		srfd_error("subgroup is null");
		goto put_handle;
	}

	object = srfd_get_object(handle->h_o);
	if (object == NULL) {
		srfd_error("object is null");
		goto put_subgroup;
	}

	switch (handle->h_type) {
	case SRFD_HANDLE_TYPE_READ:
		/* Downgrading of subgroup happens when handle is less than the subgroup's
		 * current level */
		if (srfd_label_ge(handle_current_lbl, srfd_get_subgroupLabel(subgroup, current_min_lbl))) {
			if (srfd_label_lt(handle_current_lbl, srfd_get_subgroupLabel(subgroup, current_lbl))) {
				srfd_set_subgroupLabel(subgroup, current_lbl, handle_current_lbl);
//				srfd_logDowngrade("subgroup %p has to be downgraded due to handle is downgraded", subgroup);
				srfd_current_lbl_propagateFromSubgroup(subgroup);
			}
		}
		break;
	case SRFD_HANDLE_TYPE_WRITE:
		/* Downgrading of object happens when handle is less than
		 * the object's current level */
//		srfd_info("checking if need to propogate to object");
		if (srfd_label_ge(handle_current_lbl, srfd_get_objectLabel(object, current_min_lbl))) {
			if (srfd_label_lt(handle_current_lbl, srfd_get_objectLabel(object, current_lbl))) {
//				srfd_info("propogating to object");
				srfd_set_objectLabel(object, current_lbl, handle_current_lbl);
				srfd_current_lbl_propagateFromObject(object);
			}
		}
		break;
	default:
		srfd_error("unknown handle type");
	}


	srfd_put_object(object);
put_subgroup:
	srfd_put_subgroup(subgroup);
put_handle:
	srfd_put_handle(handle);
out:
	return;
}
