/* * Self-Revocation Free Downgrading - A kernel module that applies information flow
 * control to protect system integrity without self-revocation
 * Copyright (C) 2014 Secure Systems Laboratory, Stony Brook University
 *
 * This file is part of Self-Revocation Free Downgrading.
 *
 * Self-Revocation Free Downgrading is free software: you can redistribute it
 * and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Self-Revocation Free Downgrading is distributed in the hope that it will be
 * useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Self-Revocation Free Downgrading.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

#include "srfd_security.h"
#include "srfd_debug.h"
#include "srfd_internal.h"

void srfd_current_min_lbl_updateWithNewFromHandle(struct SRFD_handle *handle) {

	if (srfd_handle_current_min_lbl_propagated(handle)) {
		srfd_info("min_lbl is already propagated");
		return;
	}

	srfd_info("handle will increase its current_min_lbl from %d to %d",
			srfd_get_handleLabelV(handle, current_min_lbl),
			srfd_get_handleLabelV(handle, new_current_min_lbl));

	srfd_handle_update_current_min_lbl(handle);

	switch (handle->h_type) {
	case SRFD_HANDLE_TYPE_READ:
		srfd_current_min_lbl_updateWithNewFromObject(handle->h_o);
		return;
	case SRFD_HANDLE_TYPE_WRITE:
		srfd_current_min_lbl_updateWithNewFromSubgroup(handle->h_sg);
		return;
	default:
		srfd_error("Not handled handle type");
		return;
	}
}


void srfd_current_min_lbl_updateWithNewFromObject(struct SRFD_object *object) {

	struct SRFD_handle *wh;

	if (srfd_object_current_min_lbl_propagated(object))
		return;

	srfd_info("object will increase its current_min_lbl from %d to %d",
		srfd_get_objectLabelV(object, current_min_lbl),
		srfd_get_objectLabelV(object, new_current_min_lbl));

	srfd_object_update_current_min_lbl(object);

	srfd_do_foreach_object_valid_wh(object, wh) {
		srfd_current_min_lbl_propagateFromHandle(wh);
	} srfd_while_foreach_object_wh(object);
}


void srfd_current_min_lbl_updateWithNewFromSubgroup(struct SRFD_subgroup *subgroup) {

	struct SRFD_handle *rh;

	if (srfd_subgroup_current_min_lbl_propagated(subgroup))
		return;

	srfd_info("subject_group %p will increase its current_min_lbl from %d to %d",
		subgroup,
		srfd_get_subgroupLabelV(subgroup, current_min_lbl),
		srfd_get_subgroupLabelV(subgroup, new_current_min_lbl));

	srfd_subgroup_update_current_min_lbl(subgroup);

	srfd_do_foreach_subgroup_valid_rh(subgroup, rh) {
		srfd_current_min_lbl_propagateFromHandle(rh);
	} srfd_while_foreach_subgroup_rh(subgroup);
}



/* Determine if the subgroup current_min_lbl can be decreased to a specific target */
int srfd_current_min_lbl_subgroupCanDecrease2Target(struct SRFD_subgroup *subgroup, struct SRFD_lbl target_min_lbl) {

	struct SRFD_handle *wh;

	/* Do nothing if the subgroup already satisfied the requirement */
	if (srfd_label_le(srfd_get_subgroupLabel(subgroup, current_min_lbl), target_min_lbl))
		return 1;

	/* Cannot decrease if the min_lbl of the subgroup is greater than target
		subgroup min lbl cannot drop below target
	*/
	if (srfd_label_gt(srfd_get_subgroupLabel(subgroup, min_lbl), target_min_lbl))
		return -1;

	/* If the new_current_min_lbl is <= target, then we should have the label decreased beforei????? */
	if (srfd_label_ge(target_min_lbl, srfd_get_subgroupLabel(subgroup, new_current_min_lbl)))
		return 1;

	/* Otherwise, subgroup's current_min_lbl is greater then the target value. Check if can decrease:
		First, set new_current_min_lbl to be the target, and see if the others can also be decreased
		This is because the object itself may depend on this subgroup.
	*/
	srfd_set_subgroupLabel(subgroup, new_current_min_lbl, target_min_lbl);


	srfd_info("list of valid write handle");

	srfd_do_foreach_subgroup_valid_wh(subgroup, wh) {
		srfd_get_handle(wh);
		srfd_info("Testing if current_min_lbl of wh obj ino %lu of subgroup %p can be decreased", wh->h_o->o_ino, subgroup);
		if (srfd_current_min_lbl_handleCanDecrease2Target(wh, target_min_lbl) == -1) {
			srfd_beforebreak_foreach_subgroup_wh(subgroup);
			srfd_put_handle(wh);
			goto reset_write_handle_new_current_min_lbl;
		} else {
			srfd_info("handle with inode ino %lu can propagate label", wh->h_o->o_inode->i_ino);
			srfd_put_handle(wh);
		}
	} srfd_while_foreach_subgroup_wh(subgroup);

	srfd_info("subject_group %p can decrease its current_min_lbl from %d to %d", subgroup,
			srfd_get_subgroupLabelV(subgroup, current_min_lbl),
			srfd_get_subgroupLabelV(subgroup, new_current_min_lbl));

	return 1;

reset_write_handle_new_current_min_lbl:
	srfd_do_foreach_subgroup_valid_wh(subgroup, wh) {
		srfd_get_handle(wh);
		srfd_handle_resetNewMinToCurMin(wh);
		srfd_put_handle(wh);
	} srfd_while_foreach_subgroup_wh(subgroup);

	srfd_subgroup_resetNewMinToCurMin(subgroup);

	return -1;
}


int srfd_current_min_lbl_objectCanDecrease2Target(struct SRFD_object *object, struct SRFD_lbl target_min_lbl) {

	struct SRFD_handle *rh;

	srfd_logTEMP("object ino %lu cur min %d, min %d, target min %d", object->o_ino, srfd_get_objectLabelV(object, current_min_lbl), srfd_get_objectLabelV(object, min_lbl), SRFD_label2int(target_min_lbl));

	/* Do nothing if the object already satisfied the requirement */
	if (srfd_label_le(srfd_get_objectLabel(object, current_min_lbl), target_min_lbl))
		return 1;

	/* Cannot decrease if the min level of the object is less than target
		 cannot drop below min_lbl
	*/
	if (srfd_label_gt(srfd_get_objectLabel(object, min_lbl), target_min_lbl))
		return -1;

	/* If the new_current_min_lbl is <= target, then we should have the label decreased before */
	if (srfd_label_ge(target_min_lbl, srfd_get_objectLabel(object, new_current_min_lbl)))
		return 1;

	/* Otherwise, object's current_min_lbl is greater then the target value. Check if can be decreased:
		First, set new_current_min_lbl to be the target, and see if the others can also be decreased
		This is because the subgroup itself may depend on this object.
	*/
	srfd_set_objectLabel(object, new_current_min_lbl, target_min_lbl);

	/* Check for each of the subgroup that opened this object for reading
		if they can also decrease the min label or not */
	srfd_do_foreach_object_valid_rh(object, rh) {
		srfd_get_handle(rh);
		if (srfd_current_min_lbl_handleCanDecrease2Target(rh, target_min_lbl) == -1) {
			srfd_info("read handle cannot decrease its label");
			srfd_beforebreak_foreach_object_rh(object);
			srfd_put_handle(rh);
			goto reset_read_handle_new_current_min_lbl;
		}
		srfd_put_handle(rh);
	} srfd_while_foreach_object_rh(object);


	srfd_info("object ino %lu can increase its current_min_lbl from %d to %d", object->o_inode->i_ino,
			srfd_get_objectLabelV(object, current_min_lbl),
			srfd_get_objectLabelV(object, new_current_min_lbl));

	return 1;

reset_read_handle_new_current_min_lbl:

	srfd_do_foreach_object_valid_rh(object, rh) {
		srfd_get_handle(rh);
		srfd_handle_resetNewMinToCurMin(rh);
		srfd_put_handle(rh);
	} srfd_while_foreach_object_rh(object);

	srfd_object_resetNewMinToCurMin(object);

	return -1;

}


int srfd_current_min_lbl_handleCanDecrease2Target(struct SRFD_handle *handle, struct SRFD_lbl target_min_lbl) {
        struct SRFD_lbl required_new_current_min_lbl;

	/* Compute what is the required current_min_lbl if the upstream has a updated new_current_min_lbl */
	required_new_current_min_lbl = srfd_computeHandle_new_current_min_lbl(handle, target_min_lbl);

        /* The current level cannot satisfy the requirement when the required min lbl is higher than the current label*/
        if (srfd_label_gt(required_new_current_min_lbl, srfd_get_handleLabel(handle, current_lbl))) {
                srfd_info("The handle current label is less than the new current min lbl");
                return -1;
        }

	/* Check if propogation is required
		The required new_current_min_lbl can be less than the target min lbl if the handle is
		invulnerable, ....
		Hence, from this point onward, will simply based on what the handle has, instead of
		forcing all downstream to use target min lbl.
	*/
	
	srfd_logTEMP("required cur min %d, handle new cur min %d", SRFD_label2int(required_new_current_min_lbl), srfd_get_handleLabelV(handle, new_current_min_lbl));

	/* Try decreasing if the target is less than required */
        if (srfd_label_lt(required_new_current_min_lbl, srfd_get_handleLabel(handle, new_current_min_lbl))) {

                srfd_set_handleLabel(handle, new_current_min_lbl, required_new_current_min_lbl);

		//if (srfd_new_current_min_lbl_canPropagateDownstreamFromHandle2(handle) == -1) {
		if (srfd_current_min_lbl_canDecreaseDownstream(handle, required_new_current_min_lbl) == -1) {
			srfd_info("Failed to propagate current_min_lbl to downstream");
			srfd_handle_resetNewMinToCurMin(handle);
			return -1;
		}
		srfd_info("handle ino %lu can decrease its current_min_lbl from %d to %d", handle->h_o->o_inode->i_ino,
				srfd_get_handleLabelV(handle, current_min_lbl),
				srfd_get_handleLabelV(handle, new_current_min_lbl));

        }
        return 1;

}


int srfd_current_min_lbl_canDecreaseDownstream(struct SRFD_handle *handle, struct SRFD_lbl target_min_lbl) {

	int rv;

	switch (handle->h_type) {
	case SRFD_HANDLE_TYPE_WRITE: {
		struct SRFD_object *object = srfd_get_object(handle->h_o);	
		srfd_info("propagate from handle to object %p: new current_min_lbl %d", handle->h_o, target_min_lbl.level);
		srfd_logTEMP("propagate from handle to object %p: new current_min_lbl %d", handle->h_o, target_min_lbl.level);
		rv = srfd_current_min_lbl_objectCanDecrease2Target(handle->h_o, target_min_lbl);
		srfd_put_object(object);
		return rv;
	}
	case SRFD_HANDLE_TYPE_READ: {/* From subject to object */
		struct SRFD_subgroup *subgroup = srfd_get_subgroup(handle->h_sg);
		srfd_info("propagate from handle to subgroup %p: new current_min_lbl %d", handle->h_sg, target_min_lbl.level);
		srfd_logTEMP("propagate from handle to subgroup %p: new current_min_lbl %d", handle->h_sg, target_min_lbl.level);
		rv = srfd_current_min_lbl_subgroupCanDecrease2Target(handle->h_sg, target_min_lbl);
		srfd_put_subgroup(subgroup);
		return rv;
	}
	default:
		srfd_error("Unknown handle type");
		return -1;
	}
}


/* Determine if the subgroup current_min_lbl can be increased to a specific target */
int srfd_current_min_lbl_subgroupCanIncrease2Target(struct SRFD_subgroup *subgroup, struct SRFD_lbl target_min_lbl) {

	struct SRFD_handle *rh;

	/* Do nothing if the subgroup already satisfied the requirement */
	if (srfd_label_ge(srfd_get_subgroupLabel(subgroup, current_min_lbl), target_min_lbl))
		return 1;

	/* Cannot increase if the subgroup's current level is less then the target
		min_lbl is to prevent subgroup from downgrading too much, hence current >= min_lbl
	 */
	if (srfd_label_lt(srfd_get_subgroupLabel(subgroup, current_lbl), target_min_lbl))
		return -1;

	/* If the new_current_min_lbl is >= target, then we should have the label increased before */
	if (srfd_label_le(target_min_lbl, srfd_get_subgroupLabel(subgroup, new_current_min_lbl)))
		return 1;

	/* Otherwise, subgroup's current_min_lbl is less then the target value. Check if can increase:
		First, set new_current_min_lbl to be the target, and see if the others can also be increased
		This is because the object itself may depend on this subgroup.
	*/
	srfd_set_subgroupLabel(subgroup, new_current_min_lbl, target_min_lbl);

	srfd_do_foreach_subgroup_valid_rh(subgroup, rh) {
		srfd_info("subgroup %p reads from ino %lu, checking if can increase 2 target", subgroup, rh->h_o->o_inode->i_ino);
		srfd_get_handle(rh);
		if (srfd_current_min_lbl_handleCanIncrease2Target(rh, target_min_lbl) == -1) {
			srfd_beforebreak_foreach_subgroup_rh(subgroup);
			srfd_put_handle(rh);
			goto reset_read_handle_new_current_min_lbl;
		} else {
			srfd_info("handle with inode ino %lu can propagate label", rh->h_o->o_inode->i_ino);
			srfd_put_handle(rh);
		}
	} srfd_while_foreach_subgroup_rh(subgroup);

	srfd_info("subject_group can increase its current_min_lbl from %d to %d",
			srfd_get_subgroupLabelV(subgroup, current_min_lbl),
			srfd_get_subgroupLabelV(subgroup, new_current_min_lbl));

	return 1;

reset_read_handle_new_current_min_lbl:
	srfd_do_foreach_subgroup_valid_rh(subgroup, rh) {
		srfd_get_handle(rh);
		srfd_handle_resetNewMinToCurMin(rh);
		srfd_put_handle(rh);
	} srfd_while_foreach_subgroup_rh(subgroup);

	srfd_subgroup_resetNewMinToCurMin(subgroup);

	return -1;


}

int srfd_current_min_lbl_objectCanIncrease2Target(struct SRFD_object *object, struct SRFD_lbl target_min_lbl) {

	struct SRFD_handle *wh;

	/* Do nothing if the object already satisfied the requirement */
	if (srfd_label_ge(srfd_get_objectLabel(object, current_min_lbl), target_min_lbl))
		return 1;

	/* Cannot increase if the object's current level is less then the target */
	if (srfd_label_lt(srfd_get_objectLabel(object, current_lbl), target_min_lbl))
		return -1;

	/* If the new_current_min_lbl is >= target, then we should have the label increased before */
	if (srfd_label_le(target_min_lbl, srfd_get_objectLabel(object, new_current_min_lbl)))
		return 1;

	/* Otherwise, object's current_min_lbl is less then the target value. Check if can increase:
		First, set new_current_min_lbl to be the target, and see if the others can also be increased
		This is because the subgroup itself may depend on this object.
	*/
	srfd_set_objectLabel(object, new_current_min_lbl, target_min_lbl);

	/* Check for each of the subgroup that opened this object for writing
		if they can also increase the min label or not */
	srfd_do_foreach_object_valid_wh(object, wh) {
		srfd_get_handle(wh);
		if (srfd_current_min_lbl_handleCanIncrease2Target(wh, target_min_lbl) == -1) {
			srfd_info("write handle cannot increase its label");
			srfd_beforebreak_foreach_object_wh(object);
			srfd_put_handle(wh);
			goto reset_write_handle_new_current_min_lbl;
		}
		srfd_put_handle(wh);
	} srfd_while_foreach_object_wh(object);


	srfd_info("object can increase its current_min_lbl from %d to %d",
			srfd_get_objectLabelV(object, current_min_lbl),
			srfd_get_objectLabelV(object, new_current_min_lbl));

	return 1;

reset_write_handle_new_current_min_lbl:

	srfd_do_foreach_object_valid_wh(object, wh) {
		srfd_get_handle(wh);
		srfd_handle_resetNewMinToCurMin(wh);
		srfd_put_handle(wh);
	} srfd_while_foreach_object_wh(object);


	srfd_object_resetNewMinToCurMin(object);

	return -1;

}


int srfd_current_min_lbl_handleCanIncrease2Target(struct SRFD_handle *handle, struct SRFD_lbl target_min_lbl) {
        struct SRFD_lbl required_new_current_min_lbl;

	/* Compute what is the required current_min_lbl if the upstream has a updated new_current_min_lbl */
	required_new_current_min_lbl = srfd_computeHandle_new_current_min_lbl(handle, target_min_lbl);

        /* The current level cannot satisfy the requirement when the required min lbl is higher than the current label*/
        if (srfd_label_gt(required_new_current_min_lbl, srfd_get_handleLabel(handle, current_lbl))) {
                srfd_info("The handle current label is less than the new current min lbl");
                return -1;
        }

	/* Check if propogation is required
		The required new_current_min_lbl can be less than the target min lbl if the handle is
		invulnerable, ....
		Hence, from this point onward, will simply based on what the handle has, instead of
		forcing all downstream to use target min lbl.
	*/

	srfd_info("checking if handle of ino %lu can increase to target min_lbl %d, now is %d", handle->h_o->o_inode->i_ino, SRFD_label2int(required_new_current_min_lbl), srfd_get_handleLabelV(handle, new_current_min_lbl));

	/* Try increasing if the required is not reached yet */
        if (srfd_label_gt(required_new_current_min_lbl, srfd_get_handleLabel(handle, new_current_min_lbl))) {
                srfd_set_handleLabel(handle, new_current_min_lbl, required_new_current_min_lbl);

		//if (srfd_new_current_min_lbl_canPropagateDownstreamFromHandle2(handle) == -1) {
		if (srfd_current_min_lbl_canIncreaseDownstream(handle, required_new_current_min_lbl) == -1) {
			srfd_info("Failed to propagate current_min_lbl to downstream");
			srfd_handle_resetNewMinToCurMin(handle);
			return -1;
		}
		srfd_info("handle can increase its current_min_lbl from %d to %d",
				srfd_get_handleLabelV(handle, current_min_lbl),
				srfd_get_handleLabelV(handle, new_current_min_lbl));

        } else {
		srfd_info("no action is required to increase handle min_lbl");
	}
        return 1;

}


int srfd_current_min_lbl_canIncreaseDownstream(struct SRFD_handle *handle, struct SRFD_lbl target_min_lbl) {

	int rv;

	switch (handle->h_type) {
	case SRFD_HANDLE_TYPE_READ: {/* From subject to object */
		struct SRFD_object *object = srfd_get_object(handle->h_o);	
		srfd_info("propagate from handle to object %p: new current_min_lbl %d", handle->h_o, target_min_lbl.level);
		rv = srfd_current_min_lbl_objectCanIncrease2Target(handle->h_o, target_min_lbl);
		srfd_put_object(object);
		return rv;
	}
	case SRFD_HANDLE_TYPE_WRITE: {
		struct SRFD_subgroup *subgroup = srfd_get_subgroup(handle->h_sg);
		srfd_info("propagate from handle to subgroup %p: new current_min_lbl %d", handle->h_sg, target_min_lbl.level);
		rv = srfd_current_min_lbl_subgroupCanIncrease2Target(handle->h_sg, target_min_lbl);
		srfd_put_subgroup(subgroup);
		return rv;
	}
	default:
		srfd_error("Unknown handle type");
		return -1;
	}
}




/************************
 *  Handle 
 ************************/

struct SRFD_lbl srfd_computeHandle_new_current_min_lbl(struct SRFD_handle *handle, struct SRFD_lbl upstream_current_min_lbl) {

        if (srfd_noConstraintOnHandle_current_min_lbl(handle)) {

                return srfd_label_min_lbl();
	}
        switch (handle->h_type) {
                case SRFD_HANDLE_TYPE_READ: /* From subject to object */
                        return srfd_label_glb(srfd_label_lub(upstream_current_min_lbl,
                                        srfd_get_handleLabel(handle, discr_min_lbl)),
                                srfd_get_handleLabel(handle, invul_lbl));
                case SRFD_HANDLE_TYPE_WRITE:
                        /* If upstream is higher than the invul_lbl, label does not flow */
			
			srfd_info("upstream: %d handle invul: %d", upstream_current_min_lbl.level, srfd_get_handleLabelV(handle, invul_lbl));
                        return srfd_label_gt(upstream_current_min_lbl, srfd_get_handleLabel(handle, invul_lbl)) ?
                                srfd_get_subgroupLabel(handle->h_sg, new_current_min_lbl) : upstream_current_min_lbl;
                default:
                        srfd_error("Unknown handle type");
                        return srfd_label_max_lbl();
        }
}


/******************************************
 * Set current_min_lbl to new_current_min_lbl
 *****************************************/

void srfd_current_min_lbl_propagateFromHandle(struct SRFD_handle *handle) {

	if (srfd_handle_current_min_lbl_propagated(handle)) {
		srfd_info("min_lbl is already propagated");
		return;
	}

	srfd_info("handle will increase its current_min_lbl from %d to %d",
			srfd_get_handleLabelV(handle, current_min_lbl),
			srfd_get_handleLabelV(handle, new_current_min_lbl));

	srfd_handle_increase_current_min_lbl(handle);

	switch (handle->h_type) {
	case SRFD_HANDLE_TYPE_READ:
		srfd_current_min_lbl_propagateFromObject(handle->h_o);
		return;
	case SRFD_HANDLE_TYPE_WRITE:
		srfd_current_min_lbl_propagateFromSubgroup(handle->h_sg);
		return;
	default:
		srfd_error("Not handled handle type");
		return;
	}
}


void srfd_current_min_lbl_propagateFromObject(struct SRFD_object *object) {

	struct SRFD_handle *wh;

	if (srfd_object_current_min_lbl_propagated(object))
		return;

	srfd_info("object will increase its current_min_lbl from %d to %d",
		srfd_get_objectLabelV(object, current_min_lbl),
		srfd_get_objectLabelV(object, new_current_min_lbl));

	srfd_object_increase_current_min_lbl(object);

	srfd_do_foreach_object_valid_wh(object, wh) {
		srfd_current_min_lbl_propagateFromHandle(wh);
	} srfd_while_foreach_object_wh(object);
}


void srfd_current_min_lbl_propagateFromSubgroup(struct SRFD_subgroup *subgroup) {

	struct SRFD_handle *rh;

	if (srfd_subgroup_current_min_lbl_propagated(subgroup))
		return;

	srfd_info("subject_group %p will increase its current_min_lbl from %d to %d",
		subgroup,
		srfd_get_subgroupLabelV(subgroup, current_min_lbl),
		srfd_get_subgroupLabelV(subgroup, new_current_min_lbl));

	srfd_subgroup_increase_current_min_lbl(subgroup);

	srfd_do_foreach_subgroup_valid_rh(subgroup, rh) {
		srfd_current_min_lbl_propagateFromHandle(rh);
	} srfd_while_foreach_subgroup_rh(subgroup);
}

