/* * Self-Revocation Free Downgrading - A kernel module that applies information flow
 * control to protect system integrity without self-revocation
 * Copyright (C) 2014 Secure Systems Laboratory, Stony Brook University
 *
 * This file is part of Self-Revocation Free Downgrading.
 *
 * Self-Revocation Free Downgrading is free software: you can redistribute it
 * and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Self-Revocation Free Downgrading is distributed in the hope that it will be
 * useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Self-Revocation Free Downgrading.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

#include "srfd_security_internal.h"
#include <linux/file.h>
#include "srfd_debug.h"
#include "srfd_internal.h"

#include <linux/delay.h>

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 2, 0)
#include <linux/fdtable.h>
#endif

static DEFINE_SPINLOCK(srfd_handle_list_lock);

struct list_head srfd_handle_list[SRFD_MAX_HANDLE_HASH];
atomic_t srfd_handle_activeCount = ATOMIC_INIT(0);

#define handle2hash(handlePtr) \
	&srfd_handle_list[hash_ptr((void *)handlePtr, SRFD_HANDLE_HASH_BITS)]


void __srfd_handle_add2List(struct SRFD_handle *handle) {

	unsigned long flags;
	struct list_head *list = handle2hash(handle);

	spin_lock_irqsave(&srfd_handle_list_lock, flags);
	list_add_rcu(&handle->h_list, list);
	spin_unlock_irqrestore(&srfd_handle_list_lock, flags);
	atomic_inc(&srfd_handle_activeCount);
}


struct SRFD_handle *srfd_get_handle(struct SRFD_handle *handle) {
	if (handle != NULL) {
		atomic_inc(&handle->h_rc);
	}
	return handle;
}

static void __srfd_handle_rcuFree(struct rcu_head *rcu) {
	//srfd_alwayslog("rcuFree %p", rcu);
	struct SRFD_handle *ptr = container_of(rcu, typeof(*ptr), h_rcu);
	kfree(ptr);
	atomic_dec(&srfd_handle_activeCount);
}

void srfd_handle_free(struct SRFD_handle *handle) {
	unsigned long flags;

	spin_lock_irqsave(&srfd_handle_list_lock, flags);
	if (handle == NULL) srfd_alwayslog("handle is null");
	list_del_rcu(&handle->h_list);
	spin_unlock_irqrestore(&srfd_handle_list_lock, flags);
	call_rcu(&handle->h_rcu, __srfd_handle_rcuFree);
}


void srfd_put_handle(struct SRFD_handle *handle) {
	if (atomic_dec_and_test(&handle->h_rc)) {
		srfd_handle_free(handle);
	}
}


struct SRFD_handle *srfd_handle_alloc(void) {
	
	struct SRFD_handle *handle = kzalloc(sizeof(*handle), GFP_ATOMIC);

	if (srfd_assert1r(handle != NULL))
		return NULL;

	atomic_set(&handle->h_rc, 0);
	__srfd_handle_add2List(handle);

	return handle;
}

int srfd_handle_matchWithFile(struct SRFD_handle *handle, struct file *file) {

	int ret = 0;
        struct SRFD_object *object;

	if (srfd_assert1r(srfd_get_handle(handle) != NULL))
		goto out;

	object = srfd_get_object(handle->h_o);
	if (srfd_assert1r(object != NULL))
		goto put_handle;

	srfd_assert1(object->o_inode != NULL);
        if (file && file->f_dentry->d_inode && object &&
		(file->f_dentry->d_inode->i_ino == object->o_ino ||
	  	file->f_dentry->d_inode->i_ino == object->o_socket_ino)) {
                if ((handle->h_type == SRFD_HANDLE_TYPE_READ && (file->f_mode & FMODE_READ)) ||
                        (handle->h_type == SRFD_HANDLE_TYPE_WRITE && (file->f_mode & FMODE_WRITE))) {
                        ret = 1;
                }
        }

	srfd_put_object(object);
put_handle:
	srfd_put_handle(handle);
out:
        return ret;
}

struct SRFD_handle *srfd_handle_dupWithNewSubgroup(struct SRFD_handle *oldHandle, struct SRFD_subgroup *newSubgroup) {
	struct SRFD_handle *newHandle = NULL;
	struct SRFD_object *object;

	if (srfd_assert1r(srfd_get_handle(oldHandle) != NULL))
		goto out;

	object = srfd_get_object(oldHandle->h_o);
	if (srfd_assert1r(object != NULL))
		goto put_old_handle;

	newHandle = srfd_get_handle(srfd_handle_alloc());
	if (srfd_assert1r(newHandle != NULL))
		goto put_object;

	newHandle->h_type = oldHandle->h_type;
	newHandle->h_label = oldHandle->h_label;
	newHandle->h_canHandleErrors = oldHandle->h_canHandleErrors;
	newHandle->h_isFlowMediated = oldHandle->h_isFlowMediated;
	newHandle->h_isDelayed = oldHandle->h_isDelayed;
	newHandle->h_fd = oldHandle->h_fd;
	newHandle->h_sg = newSubgroup;
	newHandle->h_o = object;

	srfd_info("newhandle handle type is %d, old %d", newHandle->h_type, oldHandle->h_type);

	srfd_addHandle2Object(newHandle, object);
	srfd_addHandle2Subgroup(newHandle, newSubgroup);

put_object:
	srfd_put_object(object);
put_old_handle:
	srfd_put_handle(oldHandle);
out:
	return newHandle;
}


int srfd_handle_isNoLongerValid(struct SRFD_handle *handle) {

	int ret = 1;
	struct SRFD_subgroup *subgroup;
	struct files_struct *files = NULL;
	struct fdtable *fdt;
	struct file *file;
	unsigned long flags;
	struct SRFD_subpid *subpid;
	struct task_struct *task;
	int i = 0;

	if (srfd_assert1r(srfd_get_handle(handle) != NULL))
		goto out;

	if (!handle->h_isValid) {
		srfd_alwayslog("h_isValid is false for ino %lu", handle->h_o->o_ino);
		goto put_handle;
	}

	subgroup = srfd_get_subgroup(handle->h_sg);
	if (!subgroup)
		goto put_handle;

	if (srfd_assertr(srfd_subgroup_isValid(subgroup), "subgroup %p (isInitializing: %d) has no valid member!! handle: %p, ino: %lu", subgroup, subgroup->sg_initState, handle, handle->h_o->o_ino)) {
//		dump_stack();
		if (subgroup->sg_initState != SRFD_SUBGROUP_INIT_STATE_COMPLETE)
			ret = 0;
		goto put_subgroup;
	}

	read_lock_irqsave(&subgroup->sg_sp_list_lock, flags);
	if (srfd_assertr(!list_empty(&subgroup->sg_sp_list), "subgroup %p has valid member, but list is empty!", subgroup)) {
		read_unlock_irqrestore(&subgroup->sg_sp_list_lock, flags);
		goto put_subgroup;
	}
	subpid = srfd_get_subpid(list_first_entry(&subgroup->sg_sp_list, struct SRFD_subpid, sp_sg_list));
	read_unlock_irqrestore(&subgroup->sg_sp_list_lock, flags);

	task = pid_task(subpid->sp_pid, PIDTYPE_PID);
	if (srfd_assert1r(task != NULL))
		goto put_subpid;


	files = task->files; //subject->s_task->files;

	if (srfd_assert1r(files != NULL))
		goto put_subpid;

	if (files) {

		spin_lock(&files->file_lock);
		fdt = files_fdtable(files);

		if (srfd_assert1r(fdt != NULL))
			goto unlock_files;

		if ((handle->h_fd >= 0) && (handle->h_fd < fdt->max_fds)) {
			file = srfd_fget(files, handle->h_fd);
			if (file) {
				if (srfd_handle_matchWithFile(handle, file)) {
					ret = 0;
					srfd_fput(file);
					goto unlock_files;
				}
				srfd_fput(file);
			}
		}


		for (i = 0; i < fdt->max_fds; i++) {
			file = srfd_fget(files, i);

			if (!file) {
				continue;
			}
			if (srfd_handle_matchWithFile(handle, file)) {
				handle->h_fd = i;
				ret = 0;
				srfd_fput(file);
				break;
			} else {
				srfd_fput(file);
			}
		}

unlock_files:
		spin_unlock(&files->file_lock);
	}



	if (ret) {
		srfd_info("subgroup %p handle with ino %lu, mode is read? %d is nolongervalid", subgroup, handle->h_o->o_inode->i_ino, handle->h_type == SRFD_HANDLE_TYPE_READ);
	}

put_subpid:
	srfd_put_subpid(subpid);

put_subgroup:
	srfd_put_subgroup(subgroup);

put_handle:
	srfd_put_handle(handle);
out:
	return ret;
}




