/* * Self-Revocation Free Downgrading - A kernel module that applies information flow
 * control to protect system integrity without self-revocation
 * Copyright (C) 2014 Secure Systems Laboratory, Stony Brook University
 *
 * This file is part of Self-Revocation Free Downgrading.
 *
 * Self-Revocation Free Downgrading is free software: you can redistribute it
 * and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Self-Revocation Free Downgrading is distributed in the hope that it will be
 * useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Self-Revocation Free Downgrading.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

#include <linux/security.h>
#include "security_ops_finder.h"
#include "srfd_debug.h"
#include "srfd_lsm.h"
#include "srfd_label.h"

#include "srfd_security.h"
#include <linux/binfmts.h>
#include "srfd_internal.h"


#include "srfd_abstract.h"

#include "srfd_netlink.h"

#include <net/af_unix.h>

/*#include "srfd_file.h"
*/

#define CHECK_FILES 1

static struct security_operations original_security_ops;


int srfd_unix_stream_connect(struct sock *sock, struct sock *other, struct sock *newsk) {


	struct SRFD_object *object = NULL;
	struct SRFD_subgroup *subgroup;
	struct inode *inode = NULL;
	int rc = 0;

	srfd_info("This socket ino: %lu", sock->sk_socket->file->f_dentry->d_inode->i_ino);
	srfd_info("Peer socket ino: %lu", other->sk_socket->file->f_dentry->d_inode->i_ino);

	if (other && other->sk_socket && other->sk_socket->file && other->sk_socket->file->f_dentry->d_inode) {
		inode = other->sk_socket->file->f_dentry->d_inode;
		object = SRFD_GET_OBJECT_BY_INODE(inode);
		object->o_socket_ino = sock->sk_socket->file->f_dentry->d_inode->i_ino;
	}

	if (srfd_assert1r(object != NULL))
		goto original;

	subgroup = SRFD_GET_CURRENT_SUBGROUP();
	if (srfd_assert1r(subgroup != NULL))
		goto put_object;

	srfd_logViolation("object ino %ld has min lbl %d, cur lbl %d", inode->i_ino, srfd_get_objectLabelV(object, current_min_lbl), srfd_get_objectLabelV(object, current_lbl));
	rc = srfd_abstract_open(subgroup, object, MAY_READ|MAY_WRITE);


	if (rc) {
		struct dentry *d = d_find_alias(inode);
		if (d) {
			srfd_logViolation("subgroup %p opens inode %lu (%s) is denied", subgroup, object->o_ino, d->d_iname);
			dput(d);
		} else
			srfd_logViolation("subgroup %p opens inode %lu is denied", subgroup, object->o_ino);
	} else {
		srfd_logViolation("Connect is allowed. this socket inode is %lu", inode->i_ino);
		srfd_logViolation("object ino %ld has min lbl %d, cur lbl %d", inode->i_ino, srfd_get_objectLabelV(object, current_min_lbl), srfd_get_objectLabelV(object, current_lbl));
	}

//put_subgroup:
	SRFD_PUT_CURRENT_SUBGROUP(subgroup);
put_object:
	SRFD_PUT_OBJECT(object);

	if (rc) {
		srfd_alwayslog("inode connect is denied");
		return rc;
	}



original:
	while (!original_security_ops.unix_stream_connect);
	return original_security_ops.unix_stream_connect(sock, other, newsk);

}


int srfd_socket_sendmsg(struct socket *sock, struct msghdr *msg, int size) {

	struct SRFD_object *object = NULL;
	struct SRFD_subgroup *subgroup;
	struct inode *inode = NULL;
	struct SRFD_handle *handle = NULL;
	int rc = 0;
	struct sock *sk; 

	/* send will write to peer's object, recv will recv from own socket */

	if (!(sock && sock->sk && sock->sk->sk_family == AF_UNIX))
		goto original;

	subgroup = SRFD_GET_CURRENT_SUBGROUP();
	if (srfd_assert1r(subgroup != NULL))
		goto original;




	sk = unix_peer_get(sock->sk);
	if (sk) {
		if (sk->sk_socket && sk->sk_socket->file) {
			inode = sk->sk_socket->file->f_dentry->d_inode;
			handle = SRFD_GET_WRITE_HANDLE_IN_SUBGROUP_BY_INODE(subgroup, inode);
			if (handle == NULL) {
				object = SRFD_GET_OBJECT_BY_INODE(inode);
				object->o_socket_ino = sock->file->f_dentry->d_inode->i_ino;

				if (srfd_abstract_open(subgroup, object, MAY_WRITE) == 0) {
					handle = SRFD_GET_WRITE_HANDLE_IN_SUBGROUP_BY_INODE(subgroup, inode);
				}
				else {
					rc = -1;
					srfd_logViolation("socket sending is not allowed");
				}

			}
		}

		sock_put(sk);

		if (handle == NULL) {
			goto put_object;
		}

		rc = srfd_abstract_write(handle);
		SRFD_PUT_HANDLE_IN_SUBGROUP_BY_INODE(handle);

		if (rc) {
			srfd_logViolation("send not allowed");
		}
	} else {
		srfd_alwayslog("Cannot get peer sock");
		rc = -1;
	}

put_object:
	if (object)
		SRFD_PUT_OBJECT(object);
	SRFD_PUT_CURRENT_SUBGROUP(subgroup);


	if (rc) {
		srfd_alwayslog("socket send is denied");
		return rc;
	}

original:
	while (!original_security_ops.socket_sendmsg);
	return original_security_ops.socket_sendmsg(sock, msg, size);

}


int srfd_socket_recvmsg(struct socket *sock, struct msghdr *msg, int size, int flags) {


	struct SRFD_subgroup *subgroup;
	struct inode *inode = NULL;
	struct SRFD_handle *handle = NULL;
	int rc = 0;

	if (sock && sock->sk && (sock->sk->sk_family == AF_UNIX) && sock->file && sock->file->f_dentry->d_inode) {
		inode = sock->file->f_dentry->d_inode;
	} else
		goto original;

	subgroup = SRFD_GET_CURRENT_SUBGROUP();
	if (srfd_assert1r(subgroup != NULL))
		goto original;

	handle = SRFD_GET_READ_HANDLE_IN_SUBGROUP_BY_INODE(subgroup, inode);

	if (handle == NULL) {
		goto put_subgroup;
	}



	rc = srfd_abstract_read(handle);

	SRFD_PUT_HANDLE_IN_SUBGROUP_BY_INODE(handle);

	if (rc) {
		srfd_logViolation("read not allowed");
	}

put_subgroup:
	SRFD_PUT_CURRENT_SUBGROUP(subgroup);

	if (rc) {
		srfd_alwayslog("socket read is denied");
	}

original:
	while (!original_security_ops.socket_recvmsg);
	return original_security_ops.socket_recvmsg(sock, msg, size, flags);

}


int srfd_socket_accept(struct socket *sock, struct socket *newsock) {

	/* Need to propogate the label from old to new one */
	struct SRFD_object *object = NULL;
	struct SRFD_subgroup *subgroup;
	struct inode *inode = NULL;
	int rc = 0;

	srfd_alwayslog("old sock ino: %lu", sock->file->f_dentry->d_inode->i_ino);
	srfd_alwayslog("new sock ino: %lu", newsock->file->f_dentry->d_inode->i_ino);

	if (sock && sock->file && sock->file->f_dentry->d_inode) {
		inode = sock->file->f_dentry->d_inode;
		object = SRFD_GET_OBJECT_BY_INODE(inode);
	}

	if (srfd_assert1r(object != NULL))
		goto original;

	subgroup = SRFD_GET_CURRENT_SUBGROUP();
	if (srfd_assert1r(subgroup != NULL))
		goto put_object;

		srfd_logViolation("object ino %ld has min lbl %d, cur lbl %d", inode->i_ino, srfd_get_objectLabelV(object, current_min_lbl), srfd_get_objectLabelV(object, current_lbl));
	rc = srfd_abstract_open(subgroup, object, MAY_READ|MAY_WRITE);
		srfd_logViolation("object ino %ld has min lbl %d, cur lbl %d", inode->i_ino, srfd_get_objectLabelV(object, current_min_lbl), srfd_get_objectLabelV(object, current_lbl));

	if (rc) {
		struct dentry *d = d_find_alias(inode);
		if (d) {
			srfd_logViolation("subgroup %p opens inode %lu (%s) is denied", subgroup, object->o_ino, d->d_iname);
			dput(d);
		} else
			srfd_logViolation("subgroup %p opens inode %lu is denied", subgroup, object->o_ino);
	} else {
		if (sock && sock->file && sock->file->f_dentry->d_inode) {
			srfd_logViolation("inode is %lu", sock->file->f_dentry->d_inode->i_ino);
		srfd_logViolation("object ino %ld has min lbl %d, cur lbl %d", inode->i_ino, srfd_get_objectLabelV(object, current_min_lbl), srfd_get_objectLabelV(object, current_lbl));
		}
	}

//put_subgroup:
	SRFD_PUT_CURRENT_SUBGROUP(subgroup);
put_object:
	SRFD_PUT_OBJECT(object);

	if (rc) {
		srfd_alwayslog("inode accept is denied");
		return rc;
	}

original:
	while (!original_security_ops.socket_accept);
	return original_security_ops.socket_accept(sock, newsock);

}




int srfd_socket_bind(struct socket *sock, struct sockaddr *address, int addrlen) {

	struct SRFD_object *object = NULL;
	struct SRFD_subgroup *subgroup;
	struct inode *inode = NULL;
	int rc = 0;

	if (sock && sock->file && sock->file->f_dentry->d_inode) {
		inode = sock->file->f_dentry->d_inode;
		object = SRFD_GET_OBJECT_BY_INODE(inode);
	}

	if (srfd_assert1r(object != NULL))
		goto original;

	subgroup = SRFD_GET_CURRENT_SUBGROUP();
	if (srfd_assert1r(subgroup != NULL))
		goto put_object;

	rc = srfd_abstract_open(subgroup, object, MAY_READ|MAY_WRITE);

	if (rc) {
		struct dentry *d = d_find_alias(inode);
		if (d) {
			srfd_logViolation("subgroup %p opens inode %lu (%s) is denied", subgroup, object->o_ino, d->d_iname);
			dput(d);
		} else
			srfd_logViolation("subgroup %p opens inode %lu is denied", subgroup, object->o_ino);
	} else {
		if (sock && sock->file && sock->file->f_dentry->d_inode) {
			srfd_logViolation("inode is %lu", sock->file->f_dentry->d_inode->i_ino);
		}
	}

//put_subgroup:
	SRFD_PUT_CURRENT_SUBGROUP(subgroup);
put_object:
	SRFD_PUT_OBJECT(object);

	if (rc) {
		srfd_alwayslog("inode bind is denied");
		return rc;
	}

original:
	while (!original_security_ops.socket_bind);
	return original_security_ops.socket_bind(sock, address, addrlen);

}


int srfd_socket_connect(struct socket *sock, struct sockaddr *address, int addrlen) {

	struct sock *sk; 
	
	srfd_alwayslog("old sock ino: %lu", sock->file->f_dentry->d_inode->i_ino);
	
	if (sock && sock->sk) {
		sk = unix_peer_get(sock->sk);
		if (sk) {
			srfd_alwayslog("peer sock ino: %lu", sk->sk_socket->file->f_dentry->d_inode->i_ino);
			sock_put(sk);
		} else {
			srfd_alwayslog("Cannot get peer sock");
		}
	}
	return srfd_socket_bind(sock, address, addrlen);

	while (!original_security_ops.socket_connect);
	return original_security_ops.socket_connect(sock, address, addrlen);

}

//Open, Read, Write

int srfd_inode_permission(struct inode *inode, int mask) {

	struct SRFD_object *object;
	struct SRFD_subgroup *subgroup;
	int rc = 0;

	mask &= (MAY_READ|MAY_WRITE); //|MAY_EXEC|MAY_APPEND);
	if (!mask)
		goto original;

	if (inode == NULL || SRFD_IS_IGNOREDINODE(inode))
		goto original;


	object = SRFD_GET_OBJECT_BY_INODE(inode);
	if (srfd_assert1r(object != NULL))
		goto original;

	subgroup = SRFD_GET_CURRENT_SUBGROUP();
	if (srfd_assert1r(subgroup != NULL))
		goto put_object;

	rc = srfd_abstract_open(subgroup, object, mask);

	if (rc) {
		struct dentry *d = d_find_alias(inode);
		if (d) {
			srfd_logViolation("subgroup %p opens inode %lu (%s) is denied", subgroup, object->o_ino, d->d_iname);
			dput(d);
		} else
			srfd_logViolation("subgroup %p opens inode %lu is denied", subgroup, object->o_ino);
	}

//put_subgroup:
	SRFD_PUT_CURRENT_SUBGROUP(subgroup);
put_object:
	SRFD_PUT_OBJECT(object);

	if (rc) {
		srfd_alwayslog("inode access is denied");
		return rc;
	}

original:
	while (!original_security_ops.inode_permission);
	return original_security_ops.inode_permission(inode, mask);
}


#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 5, 0)

int srfd_file_open(struct file *f, const struct cred *cred) {


	struct SRFD_object *object;
	struct SRFD_subgroup *subgroup;
	int mask = 0;
	int rc = 0;


	if ((f->f_mode & (FMODE_READ|FMODE_WRITE)) == 0)
		goto original;

	object = SRFD_GET_OBJECT_BY_INODE(f->f_dentry->d_inode);
	if (srfd_assert1r(object != NULL))
		goto original;

	subgroup = SRFD_GET_CURRENT_SUBGROUP();
	if (srfd_assert1r(subgroup != NULL))
		goto put_object;



	if ((f->f_mode & FMODE_READ)) {
		mask |= MAY_READ;
		srfd_info("read");
	}
	if ((f->f_mode & FMODE_WRITE)) {
		mask |= MAY_WRITE;
		srfd_info("write");
	}

	srfd_info("subgroup %p opens %s, mode: %d", subgroup, f->f_dentry->d_iname, f->f_mode);

	rc = srfd_abstract_open(subgroup, object, mask);

	if (rc) {
		srfd_logViolation("subgroup %p opens ino %lu (%s) is denied", subgroup, object->o_ino, f->f_dentry->d_iname);
	}


	srfd_info("subgroup %p opens %s gives %d", subgroup, f->f_dentry->d_iname, rc);

//put_subgroup:
	SRFD_PUT_CURRENT_SUBGROUP(subgroup);
put_object:
	SRFD_PUT_OBJECT(object);


original:

	if (rc) {
		srfd_alwayslog("opening of %s is denied", f->f_dentry->d_iname);
		return rc;
	}

	while (!original_security_ops.file_open);
        return original_security_ops.file_open(f, cred);
}
#endif


void srfd_d_instantiate(struct dentry *dentry, struct inode *inode)
{
	struct SRFD_object *object;
	struct SRFD_subgroup *subgroup;
	int rc;

	if (inode == NULL)
		goto original;

	if (S_ISFIFO(inode->i_mode)) {

		object = SRFD_GET_OBJECT_BY_INODE(inode);
		if (srfd_assert1r(object != NULL))
			goto original;

		subgroup = SRFD_GET_CURRENT_SUBGROUP();
		if (srfd_assert1r(subgroup != NULL))
			goto put_object;

		srfd_info("Pipe created, has ino: %lu", inode->i_ino);

		rc = srfd_abstract_open(subgroup, object, MAY_READ|MAY_WRITE);
		srfd_assertr(rc == 0, "Open for pipe is denied");

//put_subgroup:
		SRFD_PUT_CURRENT_SUBGROUP(subgroup);
put_object:
		SRFD_PUT_OBJECT(object);
	}


original:
	while (!original_security_ops.d_instantiate);
	original_security_ops.d_instantiate(dentry, inode);

	return;
}



int srfd_file_permission (struct file *file, int mask) {

	int rc = 0;

	struct SRFD_object *object;
	struct SRFD_subgroup *subgroup;
	struct SRFD_handle *handle = NULL;


	object = SRFD_GET_OBJECT_BY_INODE(file->f_dentry->d_inode);
	if (srfd_assert1r(object != NULL))
		goto original;

	subgroup = SRFD_GET_CURRENT_SUBGROUP();
	if (srfd_assert1r(subgroup != NULL))
		goto put_object;


	//If reading directory, no downgrading
	if (!S_ISDIR(file->f_dentry->d_inode->i_mode) && (mask & MAY_READ)) {

		if (srfd_label_ge(srfd_get_objectLabel(object, current_lbl), srfd_get_subgroupLabel(subgroup, current_lbl)))
			goto check_write;

		handle = SRFD_GET_READ_HANDLE_IN_SUBGROUP_BY_INODE(subgroup, file->f_dentry->d_inode);


#if SRFD_POLICY == SRFD_POLICY_LOMAC_NOCONSTRAINT
		if (handle == NULL) {
			if (srfd_abstract_open(subgroup, object, mask) == 0) {
				handle = SRFD_GET_READ_HANDLE_IN_SUBGROUP_BY_INODE(subgroup, file->f_dentry->d_inode);
			}
			else {
				rc = -1;
				srfd_logViolation("open for reading is not allowed");
				goto put_subgroup;
			}
		}
#endif
		if (handle == NULL)
			goto check_write;

		rc = srfd_abstract_read(handle);
		SRFD_PUT_HANDLE_IN_SUBGROUP_BY_INODE(handle);

		if (rc) {
			srfd_logViolation("read not allowed");
			goto put_subgroup;
		}
	}
	
check_write:



	if (mask & MAY_WRITE) {


		if (srfd_label_ge(srfd_get_subgroupLabel(subgroup, current_lbl), srfd_get_objectLabel(object, current_lbl)))
			goto put_subgroup;


		handle = SRFD_GET_WRITE_HANDLE_IN_SUBGROUP_BY_INODE(subgroup, file->f_dentry->d_inode);

#if SRFD_POLICY == SRFD_POLICY_LOMAC_NOCONSTRAINT
		if (handle == NULL) {
			if (srfd_abstract_open(subgroup, object, MAY_WRITE) == 0) {
				handle = SRFD_GET_WRITE_HANDLE_IN_SUBGROUP_BY_INODE(subgroup, file->f_dentry->d_inode);
			}
			else {
				rc = -1;
				srfd_logViolation("open for writing is not allowed");
				goto put_subgroup;
			}
		}
#endif

		if (handle == NULL)
			goto put_subgroup;
		rc = srfd_abstract_write(handle);
		SRFD_PUT_HANDLE_IN_SUBGROUP_BY_INODE(handle);

		if (rc) {
			srfd_logViolation("write not allowed");
		}

	}


put_subgroup:
	SRFD_PUT_CURRENT_SUBGROUP(subgroup);
put_object:
	SRFD_PUT_OBJECT(object);

original:
	if (rc) {
		srfd_logViolation("write not allowed");
		return rc;
	}

	while (!original_security_ops.file_permission);
	rc = original_security_ops.file_permission(file, mask);


	return rc;
}



void srfd_inode_free_security(struct inode *inode) {

	struct SRFD_object *object = srfd_object_findAndGetByInode(inode);
	if (object != NULL) {
		srfd_object_markAsInvalid(object);
		srfd_put_object(object);
		srfd_put_object(object);
	}

	while (!original_security_ops.inode_free_security);
	return original_security_ops.inode_free_security(inode);
}






int srfd_task_create(unsigned long clone_flags) {

	int rc = 0;
	struct SRFD_credsec *credsec;
	struct SRFD_subgroup *subgroup;

	credsec = srfd_credsec_findAndGetByCred(current->cred);
	if (srfd_assert1r(credsec != NULL))
		goto original;

	subgroup = srfd_get_subgroup(credsec->cs_sg);
	if (srfd_assert1r(subgroup != NULL))
		goto put_credsec;

        if (clone_flags == (CLONE_CHILD_CLEARTID|CLONE_CHILD_SETTID|SIGCHLD)) {
		subgroup->sg_isNewGroup = 1;
	} else {
		subgroup->sg_isNewGroup = 0;
	}

	srfd_put_subgroup(subgroup);
put_credsec:
	srfd_put_credsec(credsec);
original:
        while (!original_security_ops.task_create);
        rc = original_security_ops.task_create(clone_flags);
        return rc;
}







int srfd_cred_prepare(struct cred *new, const struct cred *old, gfp_t gfp) {

	struct SRFD_subgroup *old_subgroup, *new_subgroup;
	struct SRFD_credsec *old_credsec = srfd_credsec_findAndGetByCred(old);
	struct SRFD_credsec *new_credsec = srfd_credsec_findByCred(new);
	struct SRFD_handle *handle, *newHandle;
	struct SRFD_subpid *subpid;


	if (srfd_assert1r(old_credsec != NULL))
		goto original;

	old_subgroup = old_credsec->cs_sg;

	if (!srfd_subgroup_isValid(old_subgroup)) {
		if (srfd_assertr(old == current_cred(), "subgroup has no member, and cannot attach member now!"))
			goto next;
		subpid = srfd_subpid_findAndGet(task_pid(current));
		if (subpid == NULL) {
			subpid = srfd_get_subpid(srfd_subpid_allocAndGet(task_pid(current)));
			srfd_info("creating subpid...");
			if (srfd_assert1r(subpid != NULL))
				goto next;
		}
		srfd_addSubpid2Subgroup(subpid, old_subgroup);
		srfd_put_subpid(subpid);
	}

next:
	if (srfd_assert1r(old_subgroup != NULL))
		goto put_old_credsec;

	srfd_assert1(new_credsec == NULL);
	new_credsec = srfd_credsec_allocAndGet(new);

	if (srfd_assert1r(new_credsec != NULL))
		goto put_old_credsec;


	if (old_subgroup->sg_isNewGroup) {
		old_subgroup->sg_isNewGroup = 0;
                new_subgroup = srfd_subgroup_dup(old_subgroup);
		
		if (srfd_assert1r(new_subgroup != NULL))
			goto put_new_credsec;

		new_subgroup->sg_initState = SRFD_SUBGROUP_INIT_STATE_WAITCRED;
		srfd_info("old subgroup: %p, new subgroup: %p", old_subgroup, new_subgroup);

		srfd_addCredsec2Subgroup(new_credsec, new_subgroup);
		new_credsec->cs_parent_cs = srfd_get_credsec(old_credsec);


		// For pipe to work
		srfd_do_foreach_subgroup_valid_rh(old_subgroup, handle) {
			newHandle = srfd_handle_dupWithNewSubgroup(handle, new_subgroup);
		} srfd_while_foreach_subgroup_rh(old_subgroup);


		srfd_do_foreach_subgroup_valid_wh(old_subgroup, handle) {
			newHandle = srfd_handle_dupWithNewSubgroup(handle, new_subgroup);
		} srfd_while_foreach_subgroup_wh(old_subgroup);

		srfd_info("new_subgroup %p rc is %d", new_subgroup, atomic_read(&new_subgroup->sg_rc));

	} else {
		srfd_addCredsec2Subgroup(new_credsec, old_subgroup);
	}




put_old_credsec:
	srfd_put_credsec(old_credsec);

original:
	while (!original_security_ops.cred_prepare);
	return original_security_ops.cred_prepare(new, old, gfp);

put_new_credsec:
	srfd_put_credsec(new_credsec);
	goto put_old_credsec;

}


void srfd_cred_free(struct cred *cred) {

	struct SRFD_subgroup *subgroup;
	struct SRFD_credsec *credsec = srfd_credsec_findByCred(cred);

	if (srfd_assert1r(credsec != NULL))
		goto original;

	subgroup = srfd_get_subgroup(credsec->cs_sg);
	srfd_removeCredsecFromSubgroup(credsec);
	srfd_put_subgroup(subgroup);

	if (credsec->cs_parent_cs != NULL) {
		srfd_put_credsec(credsec->cs_parent_cs);
		credsec->cs_parent_cs = NULL;
	}

	srfd_put_credsec(credsec);

original:
	while (!original_security_ops.cred_free);
	original_security_ops.cred_free(cred);
}


static int srfd_bprm_check_security(struct linux_binprm *bprm) {

	struct SRFD_subgroup *subgroup;
	struct SRFD_object *object;
	int rv = 0;

	subgroup = SRFD_GET_CURRENT_SUBGROUP();
	if (srfd_assert1r(subgroup != NULL))
		goto original;

	object = SRFD_GET_OBJECT_BY_INODE(bprm->file->f_dentry->d_inode);
	if (srfd_assert1r(object != NULL))
		goto put_subgroup;

	rv = srfd_abstract_open(subgroup, object, MAY_READ);
	if (rv) {
		srfd_logViolation("============ cannot exec object %s", bprm->file->f_dentry->d_iname);
	} else {
		srfd_info("============ can exec object %s", bprm->file->f_dentry->d_iname);
	}

	SRFD_PUT_OBJECT(object);

put_subgroup:
	SRFD_PUT_CURRENT_SUBGROUP(subgroup);

original:

	if (rv)
		return rv;

        while (!original_security_ops.bprm_check_security);
	return original_security_ops.bprm_check_security(bprm);


}



static void srfd_bprm_committing_creds(struct linux_binprm *bprm) {

	struct SRFD_subgroup *origSubgroup;
	struct SRFD_credsec *new_credsec;
	struct SRFD_object *object;

	origSubgroup = SRFD_GET_CURRENT_SUBGROUP();
	if (srfd_assert1r(origSubgroup != NULL))
		goto original;


	srfd_info("executing program %s is added to subgroup %p", bprm->file->f_dentry->d_iname, origSubgroup);
	srfd_printSubgroupInfo2(srfd_logTEMP, origSubgroup);

	new_credsec = srfd_credsec_findAndGetByCred(bprm->cred);
	if (srfd_assert1r(new_credsec != NULL))
		goto put_origsubgroup;

	//Just add the new cred to the original subgroup
	srfd_addCredsec2Subgroup(new_credsec, origSubgroup);

	object = SRFD_GET_OBJECT_BY_INODE(bprm->file->f_dentry->d_inode);
	if (srfd_assert1r(object != NULL))
		goto put_newcredsec;

	srfd_alwayslog("Executing an image with level %d, current subgroup level %d", srfd_get_subgroupLabelV(origSubgroup, current_lbl), srfd_get_objectLabelV(object, current_lbl));
	if (srfd_label_gt(srfd_get_subgroupLabel(origSubgroup, current_lbl), srfd_get_objectLabel(object, current_lbl))) {
		srfd_alwayslog("Executing an image with level lower than current subgroup level");
		srfd_set_subgroupLabel(origSubgroup, current_lbl, srfd_get_objectLabel(object, current_lbl));
		srfd_current_lbl_propagateFromSubgroup(origSubgroup);
	}

	SRFD_PUT_OBJECT(object);

put_newcredsec:
	srfd_put_credsec(new_credsec);

put_origsubgroup:
	SRFD_PUT_CURRENT_SUBGROUP(origSubgroup);

original:

        while (!original_security_ops.bprm_committing_creds);
	original_security_ops.bprm_committing_creds(bprm);

}



static int srfd_path_unlink(struct path *dir, struct dentry *dentry) {

	char *filePath = NULL, *filePathBuf;
	struct SRFD_subgroup *subgroup;
	struct SRFD_object *object;
	int rv = 0;

	goto freeFilePath;

	filePath = srfd_convertDentry2FullPath(dentry, &filePathBuf);
	srfd_logDowngrade("============= Unlink is called for file %s============", filePath);

	if (filePath && strncmp(filePath, "/home/ubuntu/", strlen("/home/ubuntu/")) == 0) {
	
		subgroup = SRFD_GET_CURRENT_SUBGROUP();
		if (srfd_assert1r(subgroup != NULL))
			goto freeFilePath;

		object = SRFD_GET_OBJECT_BY_INODE(dir->dentry->d_inode);
		if (srfd_assert1r(object != NULL))
			goto put_subgroup;

		//Directory: read and write
		//To propogate label? (which label?)
		rv = srfd_abstract_open(subgroup, object, MAY_WRITE);
		srfd_logDowngrade("subgroup %p is allowed to open object ino %lu for writing", subgroup, object->o_ino);
		srfd_printSubgroupInfo2(srfd_logDowngrade, subgroup);
		srfd_printObjectInfo2(srfd_logDowngrade, object);

		if (rv == 0) {
			srfd_assert1(srfd_label_ge(srfd_get_objectLabel(object, min_lbl),
				srfd_get_subgroupLabel(subgroup, current_lbl)));
			srfd_logDowngrade("subgroup lbl: %d, dir lbl: %d", srfd_get_subgroupLabelV(subgroup, current_lbl), srfd_get_objectLabelV(object, current_lbl));
			if (srfd_label_gt(srfd_get_objectLabel(object, current_lbl), 
				srfd_get_subgroupLabel(subgroup, current_lbl))) {
				srfd_set_objectLabel(object, current_lbl, srfd_get_subgroupLabel(subgroup, current_lbl));
				srfd_current_lbl_propagateFromObject(object);
			}

		}

//put_object:
		SRFD_PUT_OBJECT(object);
put_subgroup:
		SRFD_PUT_CURRENT_SUBGROUP(subgroup);
	}


freeFilePath:
	if (filePath) srfd_freeDentry2FullPath(filePathBuf);

	if (rv)
		return rv;

        while (!original_security_ops.path_unlink);
	return original_security_ops.path_unlink(dir, dentry);
}


static int srfd_path_rename(struct path *old_dir, struct dentry *old_dentry,
			struct path *new_dir, struct dentry *new_dentry) {

	char *oldPath, *oldPathBuf;
	char *newPath, *newPathBuf;

	struct SRFD_object *fromDirObject, *toDirObject;
	struct SRFD_subgroup *subgroup;

	int rv = 0;

	subgroup = SRFD_GET_CURRENT_SUBGROUP();
	if (srfd_assert1r(subgroup != NULL))
		goto original;


	fromDirObject = SRFD_GET_OBJECT_BY_INODE(old_dir->dentry->d_inode);
	if (srfd_assert1r(fromDirObject != NULL))
		goto put_subgroup;

	toDirObject = SRFD_GET_OBJECT_BY_INODE(new_dir->dentry->d_inode);
	if (srfd_assert1r(toDirObject != NULL))
		goto put_fromDirObject;


	oldPath = srfd_convertDentry2FullPath(old_dentry, &oldPathBuf);
	newPath = srfd_convertDentry2FullPath(new_dentry, &newPathBuf);

	srfd_info("===== Renaming from %s to %s", oldPath, newPath);

	if (srfd_abstract_open(subgroup, fromDirObject, MAY_WRITE) || srfd_abstract_open(subgroup, toDirObject, MAY_WRITE)) {
		rv = -1;
		srfd_logViolation("subgroup %p is not allowed to rename object ino %lu", subgroup, old_dentry->d_inode->i_ino);
	} else {
		srfd_info("subgroup lbl: %d, rename from dir lbl: %d, to dir lbl %d", srfd_get_subgroupLabelV(subgroup, current_lbl), 
				srfd_get_objectLabelV(fromDirObject, current_lbl), srfd_get_objectLabelV(toDirObject, current_lbl));
		if (srfd_label_gt(srfd_get_objectLabel(fromDirObject, current_lbl), srfd_get_subgroupLabel(subgroup, current_lbl))) {
			srfd_set_objectLabel(fromDirObject, current_lbl, srfd_get_subgroupLabel(subgroup, current_lbl));
			srfd_current_lbl_propagateFromObject(fromDirObject);
		}
		if (srfd_label_gt(srfd_get_objectLabel(toDirObject, current_lbl), srfd_get_subgroupLabel(subgroup, current_lbl))) {
			srfd_set_objectLabel(toDirObject, current_lbl, srfd_get_subgroupLabel(subgroup, current_lbl));
			srfd_current_lbl_propagateFromObject(toDirObject);
		}
	}

	if (oldPath) srfd_freeDentry2FullPath(oldPathBuf);
	if (newPath) srfd_freeDentry2FullPath(newPathBuf);

	SRFD_PUT_OBJECT(toDirObject);

put_fromDirObject:
	SRFD_PUT_OBJECT(fromDirObject);

put_subgroup:
	SRFD_PUT_CURRENT_SUBGROUP(subgroup);

	if (rv) return rv;

original:
	while (!original_security_ops.path_rename);
	return original_security_ops.path_rename(old_dir, old_dentry, new_dir, new_dentry);
}



#define swap_security_ops(op)                                           \
        original_security_ops.op = ops->op; smp_wmb(); ops->op = srfd_##op;

void srfd_update_security_ops(struct security_operations *ops)
{


	swap_security_ops(bprm_committing_creds);
	swap_security_ops(cred_prepare);
	swap_security_ops(cred_free);
	swap_security_ops(task_create);


#ifdef SRFD_POLICY_INTERCEPT_OPEN
	swap_security_ops(inode_permission);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 5, 0)
	swap_security_ops(file_open);
#endif
#endif

	swap_security_ops(d_instantiate);

	swap_security_ops(file_permission);
	swap_security_ops(socket_recvmsg);
	swap_security_ops(socket_sendmsg);

	swap_security_ops(inode_free_security);

	swap_security_ops(path_rename);
	swap_security_ops(path_unlink);

	swap_security_ops(bprm_check_security);
	swap_security_ops(socket_bind);
	swap_security_ops(socket_accept);
	swap_security_ops(socket_connect);
	swap_security_ops(unix_stream_connect);

	return;
}

#undef swap_security_ops


#define restore_security_ops(op)                                           \
	smp_wmb(); ops->op = original_security_ops.op;

void srfd_restore_security_ops(struct security_operations *ops) {



	restore_security_ops(bprm_committing_creds);
	restore_security_ops(cred_prepare);
	restore_security_ops(cred_free);
	restore_security_ops(task_create);


	restore_security_ops(file_permission);
	restore_security_ops(socket_recvmsg);
	restore_security_ops(socket_sendmsg);

#ifdef SRFD_POLICY_INTERCEPT_OPEN
	restore_security_ops(inode_permission);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 5, 0)
	restore_security_ops(file_open);
#endif
#endif
	restore_security_ops(d_instantiate);

	restore_security_ops(inode_free_security);

	restore_security_ops(path_rename);
	restore_security_ops(path_unlink);

	restore_security_ops(bprm_check_security);
	restore_security_ops(socket_bind);
	restore_security_ops(socket_accept);
	restore_security_ops(socket_connect);
	restore_security_ops(unix_stream_connect);
	return;
}







