/* * Self-Revocation Free Downgrading - A kernel module that applies information flow
 * control to protect system integrity without self-revocation
 * Copyright (C) 2014 Secure Systems Laboratory, Stony Brook University
 *
 * This file is part of Self-Revocation Free Downgrading.
 *
 * Self-Revocation Free Downgrading is free software: you can redistribute it
 * and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Self-Revocation Free Downgrading is distributed in the hope that it will be
 * useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Self-Revocation Free Downgrading.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

#include <linux/security.h>
#include <linux/vmalloc.h>
#include <linux/module.h>
#include <linux/seq_file.h>
#include <asm/uaccess.h>
#include <linux/fs.h>
#include <linux/file.h>
#include <linux/namei.h>
#include <linux/dcache.h>
#include <linux/fs_struct.h>
#include <linux/mount.h>
#include <linux/kernel.h>
#include <linux/mman.h>
#include <linux/proc_fs.h>
#include <linux/pagemap.h>
#include <linux/swap.h>
#include <linux/security.h>
#include <linux/skbuff.h>
#include <linux/netlink.h>
#include <linux/xattr.h>
#include <linux/hugetlb.h>
#include <linux/ptrace.h>
#include <linux/file.h>
#include <linux/fs.h>
#include <linux/ip.h>
#include <linux/un.h>
#include <linux/sched.h>
#include <net/af_unix.h>
#include <linux/namei.h>

#include "srfd_interface.h"
#include "srfd_security.h"
#include "srfd_debug.h"

#include <linux/sched.h>



static struct dentry *srfd_dir;
static struct srfd_payload srfd_global_payload;
extern int phase;

struct srfd_buffer {
	/* Exclusive lock for this structure */
	struct mutex io_sem;
	/* Type of buffer : READ / WRITE */
	int type;
	/* Buffer for writing */
	char *rw_buf;
	/* Bytes available for writing */
	int rw_avail;
	/* Size of write buffer */
	int rwbuf_size;
};

static int srfd_interface_open(struct inode *inode, struct file *file)
{
	struct srfd_buffer *buf = kmalloc(sizeof(*buf), GFP_KERNEL);
	const int key = ((u8 *) file->f_path.dentry->d_inode->i_private)
		- ((u8 *) NULL);

	if (!buf) {
		return -ENOMEM;
	}
	mutex_init(&buf->io_sem);

	switch(key) {
		case SRFD_INTERFACEWRITE:
			buf->type = (int)SRFD_INTERFACEWRITE;
			buf->rwbuf_size = sizeof(struct srfd_payload); 
			break;
		case SRFD_INTERFACEREAD:
			buf->type = (int)SRFD_INTERFACEREAD;
			buf->rwbuf_size = sizeof(struct srfd_payload); 
			break;
		case SRFD_INTERFACEREADWRITE:
			buf->rwbuf_size = 0;  /* currently this is not supported */
			break;
	}
	buf->rw_buf = kzalloc(buf->rwbuf_size, GFP_KERNEL);
	if (!buf->rw_buf) {
		kfree(buf);
		buf->rw_avail = 0;
		return -ENOMEM;
	}
	buf->rw_avail = buf->rwbuf_size;
	file->private_data = buf;
	return 0;
}

static int srfd_interface_release(struct inode *inode, struct file *file)
{
	struct srfd_buffer *buf = file->private_data;
	if(buf->rw_buf)
		kfree(buf->rw_buf);
	buf->rw_buf = NULL;
	if(buf)
		kfree(buf);
	buf = NULL;
	file->private_data = NULL;
	return 0;
}

/* srfd_find_dentry : Utility function for finding dentry from either file descriptor
 * or from file-name.
 */

struct dentry * srfd_find_dentry(struct srfd_payload *payload)
{
	struct file *fil;
	struct nameidata nd;
#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 8, 0)
	int err;
	char *tmp;
#endif

	struct dentry *fil_dentry = NULL;
	if(payload->isfd){
		fil = fget((int)payload->fd);
		if (!fil) {		
			printk("error fget file\n");
			goto srfd_get_dentry_error;
		}
		fil_dentry = fil->f_dentry;
		if (!fil_dentry) {
			printk("NULL dentry\n");
			goto srfd_get_dentry_error;
		}
	}
	else {

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 8, 0)
		char *kpath = kmalloc(PATH_MAX, GFP_KERNEL);
		int error;

		if (kpath == NULL)
			goto srfd_get_dentry_error;


		error = strncpy_from_user(kpath, payload->user_name, PATH_MAX);
		srfd_alwayslog("getting dentry of path %s", kpath);
		srfd_alwayslog("error is  %d", error);
		if (error == 0)
			error = -ERANGE;
		if (error < 0)
			goto out;

		if (kern_path(kpath, 0, &(nd.path)))
			goto out;

		fil_dentry = nd.path.dentry;
		srfd_alwayslog("file dentry is %p", fil_dentry);
		dget(fil_dentry);

out:
		kfree(kpath);
#else

		tmp = getname(payload->user_name);
		err = PTR_ERR(tmp);

		if (!IS_ERR(tmp)) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 32)
			err = kern_path(tmp, 0, &(nd.path));
			fil_dentry = nd.path.dentry;
#else
			err = path_lookup(tmp, 0, &nd);
			fil_dentry = nd.dentry;
#endif
			putname(tmp);
			if (err) {
				goto srfd_get_dentry_error;
			}



			dget(fil_dentry);
		}
		else {
			goto srfd_get_dentry_error;
		}
#endif
	}
srfd_get_dentry_error:
	return fil_dentry;

}

static int srfd_process_interface_cmd (struct srfd_payload *payload)
{

	int err = -EINVAL;
	switch (payload->cmd) {
		case SRFD_GET_LABEL:
			{
				struct dentry * fil_dentry;
				struct SRFD_object *object;

				fil_dentry = srfd_find_dentry(payload);
				if(!fil_dentry) {
					printk("Bad File\n");
					return -EBADF;
				}
				err = srfd_get_fileLabel(fil_dentry->d_inode, 
						&payload->file_lbl);


				object = srfd_object_findAndGetByInode(fil_dentry->d_inode);
				if (object != NULL) {
					memcpy(&payload->file_lbl.obj_lbl, &object->o_label, sizeof(struct SRFD_object_lbl));
					err = 0;
					srfd_put_object(object);
				}

				dput(fil_dentry);

			}
			break;
		case SRFD_SET_LABEL:
			{
				struct dentry * fil_dentry;
				fil_dentry = srfd_find_dentry(payload);

				if(!fil_dentry) {
					printk("Bad File\n");
					return -EBADF;
				}
				printk("dentry's inode =  %d\n",(int)fil_dentry->d_inode->i_ino);
				// dump_object_label(&payload->file_lbl.obj_lbl);
				err = srfd_set_fileLabel(fil_dentry->d_inode, &payload->file_lbl, 1);
				dput(fil_dentry);
			}
			break;
		default:
			srfd_error("Unsupported command");

	}
	return err;
}

static ssize_t srfd_interface_write(struct file *file, 
		const char __user *user_buf, size_t count, loff_t *ppos)
{
	char *cp0;
	int error;
	struct srfd_buffer *buf = file->private_data;
	
	cp0 = buf->rw_buf;
	
	if (!access_ok(VERIFY_READ, user_buf, count)){
		return -EFAULT;
	}
	if (mutex_lock_interruptible(&buf->io_sem)){
		return -EINTR;
	}
	
	if (copy_from_user(cp0, user_buf, buf->rwbuf_size)) {
		cp0 = ERR_PTR(-EFAULT);
	}

	if (!IS_ERR(cp0)) {
		memcpy (&srfd_global_payload, (struct srfd_payload *)buf->rw_buf,
							buf->rwbuf_size);
		error = srfd_process_interface_cmd(&srfd_global_payload);
	}
	else {
		error = -EFAULT;
		goto skip;
	}
	
	srfd_global_payload.error = error;
	
skip:
	mutex_unlock(&buf->io_sem);
	return error;
}

static ssize_t srfd_interface_set_read(struct file *file, char __user *user_buf,
						size_t count, loff_t *ppos)
{
	return 0;
}

static ssize_t srfd_interface_read(struct file *file, char __user *user_buf, 
					size_t count, loff_t *ppos)
{
	int error = 0;
	struct srfd_buffer *buf = file->private_data;
	
	if (!access_ok(VERIFY_WRITE, user_buf, buf->rwbuf_size)){
		return -EFAULT;
	}
	if (mutex_lock_interruptible(&buf->io_sem)) {
		return -EINTR;
	}
	
	if (copy_to_user(user_buf, &srfd_global_payload, buf->rwbuf_size)) {
		error = -EFAULT;
		goto read_skip;
	}
	printk("[SRFDFS] Get Operation Returns !!! \n");
read_skip:
	mutex_unlock(&buf->io_sem);
	return error;
}


/* securityfs releated interfaces */

static const struct file_operations srfd_write_ops = {	
	.open    = srfd_interface_open,
	.release = srfd_interface_release,
	.write   = srfd_interface_write,
	.read    = srfd_interface_set_read,
};
static const struct file_operations srfd_read_ops = {	
	.open    = srfd_interface_open,
	.release = srfd_interface_release,
	.read    = srfd_interface_read,
};

static void srfdfs_remove(const char *name)                                      
{
	struct dentry *dentry;
	dentry = lookup_one_len(name, srfd_dir, strlen(name));
	if (!IS_ERR(dentry)) {
		securityfs_remove(dentry);
		dput(dentry);
	}
}

static void srfd_create_entry(const char *name, const mode_t mode,
					struct dentry *parent, const u8 key)
{
	struct dentry *dentry;
	switch (key) {
		case SRFD_INTERFACEWRITE: 
			dentry = securityfs_create_file(name, mode, parent,
					((u8 *) NULL) + key, &srfd_write_ops);
			break;
		case SRFD_INTERFACEREAD: 
			dentry = securityfs_create_file(name, mode, parent,
					((u8 *) NULL) + key, &srfd_read_ops);
			break;
	}
	srfd_alwayslog("dentry returned is %p", dentry);
}


int srfd_create_srfdfs(void)
{

        srfd_dir = securityfs_create_dir("srfd", NULL);
        srfd_create_entry("srfd_interface_set", 0200, srfd_dir,
                                                SRFD_INTERFACEWRITE);
        srfd_create_entry("srfd_interface_get", 0400, srfd_dir,
                                                SRFD_INTERFACEREAD);
        return 0;
}

void srfd_destroy_srfdfs(void)
{
        if (srfd_dir) {
                srfdfs_remove("srfd_interface_set");
                srfdfs_remove("srfd_interface_get");
                securityfs_remove(srfd_dir);
                srfd_dir = NULL;
        }
}

