/* * Self-Revocation Free Downgrading - A kernel module that applies information flow
 * control to protect system integrity without self-revocation
 * Copyright (C) 2014 Secure Systems Laboratory, Stony Brook University
 *
 * This file is part of Self-Revocation Free Downgrading.
 *
 * Self-Revocation Free Downgrading is free software: you can redistribute it
 * and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Self-Revocation Free Downgrading is distributed in the hope that it will be
 * useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Self-Revocation Free Downgrading.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

#include <linux/module.h>
#include <net/sock.h>
#include <linux/netlink.h>
#include <linux/skbuff.h>
#include "srfd_debug.h"
#include "srfd_netlink.h"


#define NETLINK_USER 31

struct sock *srfd_nl_sk = NULL;
static pid_t srfd_nl_daemonPid = 0;

static void srfd_nl_recv_msg(struct sk_buff *skb);



#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 6, 0)
struct netlink_kernel_cfg cfg = {
	.input = srfd_nl_recv_msg,
};
#endif


struct sock *srfd_nl_getSk(void) {
	
	if (srfd_nl_sk != NULL)
		return srfd_nl_sk;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 6, 0)
	srfd_nl_sk = netlink_kernel_create(&init_net, NETLINK_USER, &cfg);
#else
	srfd_nl_sk = netlink_kernel_create(&init_net, NETLINK_USER, 0, srfd_nl_recv_msg, NULL, THIS_MODULE);
#endif
	if (!srfd_nl_sk) {
		srfd_error("Error creating netlink socket");
	}
	srfd_info("netlink socket created");

	return srfd_nl_sk;
}

void srfd_nl_init(void) {

	if (srfd_nl_getSk() == NULL) {
		srfd_error("Failed to init netlink");
	} else {
		srfd_info("Netlink init done");
	}
}

void srfd_nl_fin(void) {
	if (srfd_nl_sk != NULL) {
		netlink_kernel_release(srfd_nl_sk);
	}
	srfd_info("Netlink fin done");
}


void srfd_nl_send_msg(void *message, size_t msg_size) {

	struct sock *nlsk = srfd_nl_getSk();
	if (nlsk != NULL && srfd_nl_daemonPid != 0) {

		struct sk_buff *skb_out = nlmsg_new(msg_size, 0);
		struct nlmsghdr *nlh;
		int res;

		if (!skb_out) {
			srfd_error("Failed to allocate new skb");
			return;
		}

		nlh = nlmsg_put(skb_out, 0, 0, NLMSG_DONE, msg_size, 0);
		NETLINK_CB(skb_out).dst_group = 0; /* not in mcast group */
		memcpy(nlmsg_data(nlh), message, msg_size);

		res = nlmsg_unicast(srfd_nl_sk, skb_out, srfd_nl_daemonPid);

		if (res == -ECONNREFUSED) {
			srfd_nl_daemonPid = 0;
			srfd_alwayslog("daemon stopped receiving messages");
			return;
		}

		if (res != 0)
			srfd_alwayslog("Unexpected error message from send: returned %d", res);

	}
}


void srfd_nl_log2daemon2(char *message, size_t len) {
	if (message != NULL)
		srfd_nl_send_msg(message, len);
}



static void srfd_nl_recv_msg(struct sk_buff *skb) {

	struct nlmsghdr *nlh;
	int pid;
	int msg_size;
	char *msg = "Hello from kernel";

	srfd_alwayslog("message received");
	nlh = (struct nlmsghdr *)skb->data;
	srfd_nl_daemonPid = pid = nlh->nlmsg_pid; /*pid of sending process */
	srfd_alwayslog("payload size is %d", nlh->nlmsg_len);
	srfd_alwayslog("message is from pid: %d, payload: \"%s\"", pid, (char *)nlmsg_data(nlh));


	msg_size = strlen(msg);

	srfd_nl_send_msg(msg, msg_size);

	return;
}



