/* * Self-Revocation Free Downgrading - A kernel module that applies information flow
 * control to protect system integrity without self-revocation
 * Copyright (C) 2014 Secure Systems Laboratory, Stony Brook University
 *
 * This file is part of Self-Revocation Free Downgrading.
 *
 * Self-Revocation Free Downgrading is free software: you can redistribute it
 * and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Self-Revocation Free Downgrading is distributed in the hope that it will be
 * useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Self-Revocation Free Downgrading.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

#ifndef __SRFD_NETLINK_H__
#define __SRFD_NETLINK_H__

/* For using netlink to communicate with python daemon process at the user level.
   It is currently used to send messages about objects getting downgraded.
   When an object is downgraded, the user-level daemon may move the file to
   a shadow directory such that these files won't make next application-run to
   be downgraded due to reading of non-data files.
*/


#include <net/sock.h>
#include <linux/string.h>
#include <linux/mm.h>
#include "srfd_label.h"

void srfd_nl_fin(void);
void srfd_nl_init(void);
void srfd_nl_log2daemon2(char *message, size_t len);
void srfd_nl_send_msg(void *message, size_t msg_size);


#define srfd_nl_log2daemon(_msg) do { if (_msg != NULL) srfd_nl_log2daemon2(_msg, strlen(_msg)); } while (0)



#define srfd_nl_message(type, format, args...) srfd_nl_ ##type(type, format, ##args)


#define srfd_nl_srfd_log srfd_nl_nothing
#define srfd_nl_srfd_downgrade srfd_nl_nothing
#define srfd_nl_srfd_temp srfd_nl_nothing
#define srfd_nl_srfd_objectDowngraded srfd_nl_alwaysSend


#define srfd_nl_nothing(type, format, args...)

#define srfd_nl_alwaysSend(type, format, args...) \
	do { \
		char *buf = (char *)__get_free_page(GFP_KERNEL); \
		if (!IS_ERR(buf)) { \
			int len = snprintf(buf+4, PAGE_SIZE, format, ##args); \
			*(int *)buf = type; \
			srfd_nl_log2daemon2(buf, len+4); \
			free_page((long unsigned int)buf); \
		} \
	} while (0)

#define srfd_nl_sendPkt(_pktptr) \
	do { \
		srfd_nl_send_msg(_pktptr, sizeof(*_pktptr)); \
	} while (0)


#define srfd_objectDowngraded(_object, toLvl) \
	do { \
		char *fullPath, *buffer; \
		srfd_nl_prepare_packet(objectDowngraded, pktptr); \
		fullPath = srfd_convertInode2FullPath(_object->o_inode, &buffer); \
		if (fullPath) strncpy(pktptr->path, fullPath, sizeof(pktptr->path)); \
		else pktptr->path[0] = '\0'; \
		pktptr->toLevel = SRFD_label2int(toLvl); \
		srfd_alwayslog("OBJECT DOWNGRADED %s", fullPath); \
		srfd_nl_sendPkt(pktptr); \
		if (fullPath) srfd_freeInode2FullPath(buffer); \
		srfd_nl_free_packet(pktptr); \
	} while (0)



/* Message type identifier, the python netlink daemon must share the same id.
   To define a new message type, select an identifier, and declare how the packet
   should look like.
*/
typedef enum srfd_nl_command {
	srfd_log = 0,
	srfd_downgrade,
	srfd_temp,
	srfd_objectDowngraded
} srfd_nl_cmd;

struct srfd_nl_pkt_header {
	srfd_nl_cmd cmd;
};

#define SRFD_NL_PACKET_HEADER struct srfd_nl_pkt_header header

struct srfd_nl_pkt {
	SRFD_NL_PACKET_HEADER;
};






#define EXPAND_ARG0(...) do{}while(0)

#define EXPAND_ARG1(t1, a1) t1 a1;

#define EXPAND_ARG2(t1, a1, ...) t1 a1; \
	EXPAND_ARG1(__VA_ARGS__)


#define srfd_nl_pkt_declaration(call, argcount, ...) struct srfd_nl_pkt_ ##call { \
	SRFD_NL_PACKET_HEADER; \
	EXPAND_ARG ##argcount(__VA_ARGS__)\
}


srfd_nl_pkt_declaration(log, 1, char, message[PATH_MAX]);
srfd_nl_pkt_declaration(objectDowngraded, 2, char, path[PATH_MAX], int, toLevel);



/* Never allocate the buffer on stack!! */
#define srfd_nl_prepare_packet(_call, _pktptr) \
	struct srfd_nl_pkt_ ##_call *_pktptr = kmalloc(sizeof(*_pktptr), GFP_KERNEL); \
	if (_pktptr) { \
		_pktptr->header.cmd = srfd_ ##_call; \
	}

#define srfd_nl_free_packet(_pktptr) \
	if (_pktptr) kfree(_pktptr)


#endif /* __SRFD_NETLINK_H__ */

