/* * Self-Revocation Free Downgrading - A kernel module that applies information flow
 * control to protect system integrity without self-revocation
 * Copyright (C) 2014 Secure Systems Laboratory, Stony Brook University
 *
 * This file is part of Self-Revocation Free Downgrading.
 *
 * Self-Revocation Free Downgrading is free software: you can redistribute it
 * and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Self-Revocation Free Downgrading is distributed in the hope that it will be
 * useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Self-Revocation Free Downgrading.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

#include "srfd_security_internal.h"
#include "srfd_debug.h"

static DEFINE_SPINLOCK(srfd_object_list_lock);

struct list_head srfd_object_list[SRFD_MAX_OBJECT_HASH];

atomic_t srfd_object_activeCount = ATOMIC_INIT(0);

int srfd_object_max_search = 0;

#define inode2hash(inode) \
	&srfd_object_list[hash_ptr((inode), SRFD_OBJECT_HASH_BITS)]

struct SRFD_object_lbl srfd_object_defaultLabel = {
	.current_lbl = {7},
	.min_lbl = {1}
};


void srfd_object_fillLabelFromDisk(struct SRFD_object *object) {

	struct SRFD_file_lbl file_lbl;

	if (srfd_get_fileLabel(object->o_inode, &file_lbl)) {
		memcpy(&file_lbl.obj_lbl, &srfd_object_defaultLabel, sizeof(struct SRFD_object_lbl));
	}

	if (file_lbl.obj_lbl.min_lbl.level > file_lbl.obj_lbl.current_lbl.level) {
		srfd_info("Object ino %lu has min %d < cur %d", object->o_ino, file_lbl.obj_lbl.min_lbl.level, file_lbl.obj_lbl.current_lbl.level);
	}

	memcpy(&(object->o_label), &(file_lbl.obj_lbl), sizeof(struct SRFD_object_lbl));
	object->o_label.new_current_min_lbl = object->o_label.current_min_lbl = object->o_label.min_lbl;
	return;
}


void srfd_object_markAsInvalid(struct SRFD_object *object) {

	struct SRFD_file_lbl file_lbl;

	object->o_isValid = 0;

	if (S_ISFIFO(object->o_inode->i_mode))
		return;

	if (!object->o_inode->i_op->setxattr)
		return;

	if (srfd_get_objectLabel(object, current_lbl).level != srfd_object_defaultLabel.current_lbl.level ||
		srfd_get_objectLabel(object, min_lbl).level != srfd_object_defaultLabel.min_lbl.level) {
		int ret;

		srfd_assert1(srfd_get_fileLabel(object->o_inode, &file_lbl) == 0);
		memcpy(&file_lbl.obj_lbl, &object->o_label, sizeof(struct SRFD_object_lbl));
		ret = srfd_set_fileLabel(object->o_inode, &file_lbl, 0);

		srfd_alwayslog("saved inode label for ino %lu, ret: %d", object->o_inode->i_ino, ret);
	}
//out:
	return;
}


void __srfd_object_add2List(struct SRFD_object *ptr) {
	unsigned long flags;

	struct list_head *list = inode2hash(ptr->o_inode);
	spin_lock_irqsave(&srfd_object_list_lock, flags);
	list_add_rcu(&ptr->o_list, list);
	spin_unlock_irqrestore(&srfd_object_list_lock, flags);
	atomic_inc(&srfd_object_activeCount);

}

static void __srfd_object_rcuFree(struct rcu_head *rcu) {
	struct SRFD_object *object = container_of(rcu, typeof(*object), o_rcu);
	kfree(object);
	atomic_dec(&srfd_object_activeCount);
}

struct SRFD_object *srfd_object_findByInode(struct inode *inode) {

	struct list_head *list = inode2hash(inode);
	struct SRFD_object *object;

	while (!list->next);
	rcu_read_lock();
	list_for_each_entry_rcu(object, list, o_list) {
	if (!object->o_isValid || object->o_ino != inode->i_ino ||
		object->o_sb != inode->i_sb) {
			continue;
		}
		rcu_read_unlock();
		return object;
	}

	rcu_read_unlock();
	return NULL;
}


struct SRFD_object *srfd_object_allocAndGet(struct inode *inode) {

	struct SRFD_object *object = kzalloc(sizeof(*object), GFP_ATOMIC);

	if (!object) {
		srfd_error("Unable to allocate memory");
		return NULL;
	}

	/* Initial is 1 because there is no LSM hook for object creation? */
	atomic_set(&object->o_rc, 1);
	object->o_inode = inode;
	object->o_ino = inode->i_ino;
	object->o_sb = inode->i_sb;
	object->o_isValid = 1;
	object->o_socket_ino = 0;
	srfd_object_fillLabelFromDisk(object);
	__srfd_object_add2List(object);

	srfd_set_objectLabel(object, new_current_min_lbl, srfd_get_objectLabel(object, current_min_lbl));

	INIT_LIST_HEAD(&object->o_rh_list);
	INIT_LIST_HEAD(&object->o_wh_list);

	object->o_rh_list_lock = __SPIN_LOCK_UNLOCKED(object->o_rh_list_lock);
	object->o_wh_list_lock = __SPIN_LOCK_UNLOCKED(object->o_wh_list_lock);

	return object;

}



struct SRFD_object *srfd_get_object(struct SRFD_object *object) {
	if (object) {
		atomic_inc(&object->o_rc);
	}
	return object;
}

void srfd_put_object(struct SRFD_object *object) {
	if (atomic_dec_and_test(&object->o_rc)) {
		srfd_object_free(object);
	}
}

void srfd_object_free(struct SRFD_object *object) {

	unsigned long flags;

	if (atomic_read(&object->o_rc) > 1)
		srfd_info("Freeing object with usage count: %d > 1 %p", atomic_read(&object->o_rc), object);


	spin_lock_irqsave(&srfd_object_list_lock, flags);
	list_del_rcu(&object->o_list);
	spin_unlock_irqrestore(&srfd_object_list_lock, flags);
	call_rcu(&object->o_rcu, __srfd_object_rcuFree);

}


