/* * Self-Revocation Free Downgrading - A kernel module that applies information flow
 * control to protect system integrity without self-revocation
 * Copyright (C) 2014 Secure Systems Laboratory, Stony Brook University
 *
 * This file is part of Self-Revocation Free Downgrading.
 *
 * Self-Revocation Free Downgrading is free software: you can redistribute it
 * and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Self-Revocation Free Downgrading is distributed in the hope that it will be
 * useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Self-Revocation Free Downgrading.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

#include "srfd_security_internal.h"
#include "srfd_file.h"
#include <asm/uaccess.h>

#include <linux/delay.h>
#include "security_ops_finder.h"


struct semaphore srfd_open_sem;



int srfd_security_init(void) {

	struct list_head *super_blocks = srfd_find_super_blocks();

	struct inode *inode, *inode_n = NULL;
	struct super_block *sb, *sb_n = NULL;
	int idx;

	sema_init(&srfd_open_sem, 1);

	for (idx = 0; idx < SRFD_MAX_CREDSEC_HASH; idx++) {
		INIT_LIST_HEAD(&srfd_credsec_list[idx]);
	}

	for (idx = 0; idx < SRFD_MAX_OBJECT_HASH; idx++) {
		INIT_LIST_HEAD(&srfd_object_list[idx]);
	}

	for (idx = 0; idx < SRFD_MAX_SUBGROUP_HASH; idx++) {
		INIT_LIST_HEAD(&srfd_subgroup_list[idx]);
	}

	for (idx = 0; idx < SRFD_MAX_SUBPID_HASH; idx++) {
		INIT_LIST_HEAD(&srfd_subpid_list[idx]);
	}

	for (idx = 0; idx < SRFD_MAX_HANDLE_HASH; idx++) {
		INIT_LIST_HEAD(&srfd_handle_list[idx]);
	}


	list_for_each_entry_safe(sb, sb_n, super_blocks, s_list) {
		list_for_each_entry_safe(inode, inode_n, &sb->s_inodes, i_sb_list) {
			srfd_object_allocAndGet(inode);
		}
	}


	return 0;

}


int srfd_security_fin(void) {

	struct SRFD_object *object;
	struct SRFD_credsec *credsec;
	struct SRFD_subgroup *subgroup;
	struct SRFD_subpid *subpid;
	struct SRFD_handle *handle;


	struct list_head *list;
	int idx;
	int count;

	srfd_printStatistics();
	/* Synchronize because free will forcefully free the structure */

	count = 0;
	srfd_alwayslog("subpid to be cleaned: %d", atomic_read(&srfd_subpid_activeCount));
	rcu_read_lock();
	for (idx = 0; idx < SRFD_MAX_SUBPID_HASH; idx++) {
		list = &srfd_subpid_list[idx];
		list_for_each_entry_rcu(subpid, list, sp_list) {
			srfd_removeSubpidFromSubgroup(subpid);
			srfd_put_subpid(subpid);
			count++;
		}
	}
	rcu_read_unlock();
	srfd_alwayslog("subpid cleaned: %d", count);
	srfd_alwayslog("subpid remaining: %d", atomic_read(&srfd_subpid_activeCount));


	count = 0;
	srfd_alwayslog("credsec to be cleaned: %d", atomic_read(&srfd_credsec_activeCount));
	rcu_read_lock();
	for (idx = 0; idx < SRFD_MAX_CREDSEC_HASH; idx++) {
		list = &srfd_credsec_list[idx];
		list_for_each_entry_rcu(credsec, list, cs_list) {
			srfd_removeCredsecFromSubgroup(credsec);
			srfd_put_credsec(credsec);
			count++;
		}
	}
	rcu_read_unlock();
	srfd_alwayslog("credsec cleaned: %d", count);
	srfd_alwayslog("credsec remaining: %d", atomic_read(&srfd_credsec_activeCount));

	count = 0;
	srfd_alwayslog("handle to be cleaned: %d", atomic_read(&srfd_handle_activeCount));
	rcu_read_lock();
	for (idx = 0; idx < SRFD_MAX_HANDLE_HASH; idx++) {
		list = &srfd_handle_list[idx];
		list_for_each_entry_rcu(handle, list, h_list) {
			srfd_removeHandleFromSubgroup(handle);
			srfd_removeHandleFromObject(handle);
			srfd_put_handle(handle);
			count++;
		}
	}
	rcu_read_unlock();
	srfd_alwayslog("handle cleaned: %d", count);
	srfd_alwayslog("handle remaining: %d", atomic_read(&srfd_handle_activeCount));


	srfd_alwayslog("subgroup to be cleaned: %d", atomic_read(&srfd_subgroup_activeCount));
	rcu_read_lock();
	for (idx = 0; idx < SRFD_MAX_SUBGROUP_HASH; idx++) {
		list = &srfd_subgroup_list[idx];
		list_for_each_entry_rcu(subgroup, list, sg_list) {
			srfd_subgroup_free(subgroup);
		}
	}
	rcu_read_unlock();
	srfd_alwayslog("subgroup remaining: %d", atomic_read(&srfd_subgroup_activeCount));


	rcu_read_unlock();
	srfd_alwayslog("object to be cleaned: %d", atomic_read(&srfd_object_activeCount));
	rcu_read_lock();
	for (idx = 0; idx < SRFD_MAX_OBJECT_HASH; idx++) {
		list = &srfd_object_list[idx];
		list_for_each_entry_rcu(object, list, o_list) {
			srfd_object_markAsInvalid(object);
			srfd_object_free(object);
		}
	}
	rcu_read_unlock();
	srfd_alwayslog("object remaining: %d", atomic_read(&srfd_object_activeCount));


	while (atomic_read(&srfd_credsec_activeCount) 
		+ atomic_read(&srfd_object_activeCount)
		+ atomic_read(&srfd_subgroup_activeCount)
		+ atomic_read(&srfd_subpid_activeCount)
		+ atomic_read(&srfd_handle_activeCount)) {
		srfd_alwayslog("waiting for the pending rcu to complete...");
		synchronize_rcu();
		mdelay(100);
		srfd_printStatistics();
	}

	srfd_printStatistics();

	return 0;
}




void srfd_printStatistics(void) {

	srfd_alwayslog("credsec remaining: %d", atomic_read(&srfd_credsec_activeCount));
	srfd_alwayslog("object remaining: %d", atomic_read(&srfd_object_activeCount));
	srfd_alwayslog("subgroup remaining: %d", atomic_read(&srfd_subgroup_activeCount));
	srfd_alwayslog("subpid remaining: %d", atomic_read(&srfd_subpid_activeCount));
	srfd_alwayslog("handle remaining: %d", atomic_read(&srfd_handle_activeCount));
}


int srfd_get_fileLabel(struct inode *inode, struct SRFD_file_lbl *file_lbl) {

        struct dentry dummy_dentry;
        int ret = -1;

	memset(&dummy_dentry, 0, sizeof(struct dentry));

        dummy_dentry.d_inode = inode;
	dummy_dentry.d_sb = inode->i_sb;

        if (inode && inode->i_op && inode->i_op->getxattr) {
		if ((ret = inode->i_op->getxattr(&dummy_dentry, SRFD_XATTR_NAME, file_lbl, sizeof(*file_lbl))) == sizeof(*file_lbl)) {
			ret = 0;
		}
        }
	return ret;
}

int srfd_set_fileLabel(struct inode *inode, struct SRFD_file_lbl *file_lbl, int updateInRam) {

        struct dentry dummy_dentry;
        int ret = -1;

        dummy_dentry.d_inode = inode;
	dummy_dentry.d_sb = inode->i_sb;

        if (inode->i_op->setxattr) {

                ret = inode->i_op->setxattr(&dummy_dentry, SRFD_XATTR_NAME, file_lbl, sizeof(struct SRFD_file_lbl), 0);
		srfd_alwayslog("saving object with current lbl: %d, min lbl: %d, ret: %d", file_lbl->obj_lbl.current_lbl.level, file_lbl->obj_lbl.min_lbl.level, ret);

		if (updateInRam) {
			struct SRFD_object *object = srfd_object_findAndGetByInode(inode);
			if (object != NULL) {
				object->o_label = file_lbl->obj_lbl;
				srfd_put_object(object);
			}


		}

        }
	return ret;


}




