/* * Self-Revocation Free Downgrading - A kernel module that applies information flow
 * control to protect system integrity without self-revocation
 * Copyright (C) 2014 Secure Systems Laboratory, Stony Brook University
 *
 * This file is part of Self-Revocation Free Downgrading.
 *
 * Self-Revocation Free Downgrading is free software: you can redistribute it
 * and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Self-Revocation Free Downgrading is distributed in the hope that it will be
 * useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Self-Revocation Free Downgrading.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

#ifndef __SRFD_SECURITY_H__
#define __SRFD_SECURITY_H__

/* This file defines the core structures used by SRFD. It includes
   - SRFD_subgroup: all threads (sharing the same memory) are 
     in the same subgroup

   - SRFD_subpid: since there are no more hooks to track when a
     thread is no longer valid, pointers from existing structures may
     not be obsoleted without SRFD's knowledge. Hence SRFD maintains a
     pointer to subpid, which allows SRFD to first check if a process
     has terminated. This is particularly used for ensuring the file
     descriptor table is valid before accessing it.

   - SRFD_object: store the SRFD specific information about a particular
     inode.

   - SRFD_cred: LSM provides no hook to track task creation/destruction.
     It does, however, provides hooks to track cred creation/destruction.
     SRFD relies on this to indirectly track when a subgroup has no more
     reference, so that it can free up the structure. Furthermore, tracking
     cred is the only way to track newly cloned process before the new process
     is actually created.

  - SRFD_handle: Subgroups and objects have many to many relationship. handle
    is used to connect a subgroup and an object.
*/

#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/hash.h>
#include <linux/fs.h>
#include "srfd_label.h"
#include <linux/mm.h>

#define SRFD_POLICY_NODOWNGRADING 0
#define SRFD_POLICY_LOMAC_NOCONSTRAINT 1
#define SRFD_POLICY_LOMAC_CONSTRAINT 2

/* Select the policy to enforce
   No Dynamic Downgrading: Decide ahead of time what level to run,
     and not allow the level to be changed.
   LOMAC No Constraint: Same as LOMAC, will downgrade on read,
     fail on write
   LOMAC Constrainted = SRFD, apply constraints to limit downgrading
*/

//#define SRFD_POLICY SRFD_POLICY_NODOWNGRADING
#define SRFD_POLICY SRFD_POLICY_LOMAC_CONSTRAINT
//#define SRFD_POLICY SRFD_POLICY_LOMAC_NOCONSTRAINT

/* This allows using shell without getting shell downgraded */
#define SRFD_IS_IGNOREDINODE(_inode) (!_inode || _inode->i_ino <= 20) 


#if (SRFD_POLICY == SRFD_POLICY_LOMAC_CONSTRAINT || SRFD_POLICY == SRFD_POLICY_NODOWNGRADING)
/*
   If the policy is no dynamic downgrading (SRFD_POLICY_NODOWNGRADING), need to
   check open to see if allowed or not.
   If the policy is SRFD (SRFD_POLICY_LOMAC_CONSTRAINT), need to propogate
   min_label on open.
*/
	#define SRFD_POLICY_INTERCEPT_OPEN
#endif


#if (SRFD_POLICY == SRFD_POLICY_LOMAC_CONSTRAINT)
	#define SRFD_POLICY_ENFORCE_MIN_LABEL
#endif

int srfd_security_init(void);
int srfd_security_fin(void);

extern struct semaphore srfd_open_sem;


typedef enum {
	SRFD_HANDLE_TYPE_READ,
	SRFD_HANDLE_TYPE_WRITE,
	SRFD_HANDLE_TYPE_UNKNOWN
} SRFD_HANDLE_TYPE;


typedef enum {
	SRFD_SUBGROUP_INIT_STATE_COMPLETE,
	SRFD_SUBGROUP_INIT_STATE_DUSRFDNG,
	SRFD_SUBGROUP_INIT_STATE_WAITCRED
} SRFD_SUBGROUP_INIT_STATE;


struct SRFD_subpid {
	atomic_t			sp_rc;
	struct list_head		sp_list;

	struct pid			*sp_pid;
	struct list_head		sp_sg_list;

	struct SRFD_subgroup		*sp_sg;
	struct rcu_head 		sp_rcu;
};

struct SRFD_subpid *srfd_subpid_alloc(struct pid *pid);
struct SRFD_subpid *srfd_subpid_findByPid(struct pid *pid);
void srfd_subpid_free(struct SRFD_subpid *subpid);
int srfd_subpid_isNoLongerValid(struct SRFD_subpid *subpid);
struct SRFD_subpid *srfd_get_subpid(struct SRFD_subpid *subpid);
void srfd_put_subpid(struct SRFD_subpid *subpid);
#define srfd_subpid_allocAndGet(_pid) srfd_get_subpid(srfd_subpid_alloc(_pid))
#define srfd_subpid_findAndGet(_pid) srfd_get_subpid(srfd_subpid_findByPid(_pid))

struct SRFD_subgroup {
	
	atomic_t sg_rc;

	struct list_head 		sg_list;	/* subject group list */

	spinlock_t 			sg_rh_list_lock;
	spinlock_t 			sg_wh_list_lock;
	struct rcu_head 		sg_rcu;

	struct list_head 		sg_rh_list; 	/* read handle list */
	struct list_head 		sg_wh_list; 	/* write handle list */

	struct SRFD_sub_group_lbl 	sg_label;
	int				sg_isNewGroup;

	struct files_struct		*sg_files;
	rwlock_t			sg_files_lock;

	rwlock_t			sg_sp_list_lock;
	struct list_head		sg_sp_list;

	SRFD_SUBGROUP_INIT_STATE	sg_initState;

	atomic_t			sg_cred_rc;

};

void srfd_subgroup_free(struct SRFD_subgroup *subgroup);
struct SRFD_subgroup *srfd_get_subgroup(struct SRFD_subgroup *subgroup);
void srfd_put_subgroup(struct SRFD_subgroup *subgroup);
struct SRFD_subgroup *srfd_subgroup_alloc(void);
struct SRFD_subgroup *srfd_subgroup_dup(struct SRFD_subgroup *origSubgroup);

#define srfd_get_subgroupLabel(_subgroup, _labelName) (_subgroup->sg_label._labelName)
#define srfd_set_subgroupLabel(_subgroup, _labelName, _label) \
	do { \
		struct SRFD_lbl old_label = _subgroup->sg_label._labelName; \
		if (srfd_label_eq(old_label, _label)) break; \
		(_subgroup->sg_label._labelName = _label); \
		srfd_logDowngrade("subgroup %p " #_labelName " is changed from %d to %d", _subgroup, SRFD_label2int(old_label), SRFD_label2int(_label)); \
	} while (0)

#define srfd_get_subgroupLabelV(_subgroup, _labelName) (srfd_get_subgroupLabel(_subgroup, _labelName).level)

#define srfd_subgroup_canHandleReadError(_subgroup) srfd_get_subgroupLabel(_subgroup, can_handle_read_errors)
#define srfd_subgroup_canHandleWriteError(_subgroup) srfd_get_subgroupLabel(_subgroup, can_handle_write_errors)



struct SRFD_credsec {
	atomic_t			cs_rc;
	struct list_head		cs_list;
	struct rcu_head			cs_rcu;
	const struct cred		*cs_cred;
	struct SRFD_subgroup		*cs_sg;
	struct SRFD_credsec		*cs_parent_cs;
};

void srfd_credsec_free(struct SRFD_credsec *credsec);
struct SRFD_credsec *srfd_get_credsec(struct SRFD_credsec *credsec);
void srfd_put_credsec(struct SRFD_credsec *credsec);
struct SRFD_credsec *srfd_credsec_findByCred(const struct cred *p);
struct SRFD_credsec *srfd_credsec_allocAndGet(const struct cred *p);

#define srfd_credsec_findAndGetByCred(__cred) srfd_get_credsec(srfd_credsec_findByCred(__cred))


int srfd_addCredsec2Subgroup(struct SRFD_credsec *credsec, struct SRFD_subgroup *subgroup);
int srfd_removeCredsecFromSubgroup(struct SRFD_credsec *credsec);


struct SRFD_object {
	atomic_t		o_rc;
	unsigned long		o_ino;
	unsigned long		o_socket_ino;	/* If the object is socket, it refers to the peer's inode. o_socket_ino stores the local inode for checking */
	struct super_block	*o_sb;
	struct list_head	o_list;
	struct rcu_head		o_rcu;
	struct list_with_rc	*o_list_real_head;

	spinlock_t		o_rh_list_lock;
	spinlock_t		o_wh_list_lock;

#ifdef SRFD_USE_HANDLE_OBJECT_LOCK
	spinlock_t		o_rh_list_lock1;
	spinlock_t		o_wh_list_lock1;
#endif

	struct inode		*o_inode;
	struct SRFD_object_lbl	o_label;
	struct list_head	o_rh_list;
	struct list_head	o_wh_list;
	int			o_isValid;

};


struct SRFD_object *srfd_object_findByInode(struct inode *inode);
struct SRFD_object *srfd_object_allocAndGet(struct inode *inode);
struct SRFD_object *srfd_get_object(struct SRFD_object *object);
void srfd_put_object(struct SRFD_object *object);
void srfd_object_free(struct SRFD_object *object);

#define srfd_object_findAndGetByInode(__inode) srfd_get_object(srfd_object_findByInode(__inode))

int srfd_set_fileLabel(struct inode *inode, struct SRFD_file_lbl *file_lbl, int updateInRam);
int srfd_get_fileLabel(struct inode *inode, struct SRFD_file_lbl *file_lbl);

#define srfd_get_objectLabel(_object, _labelName) (_object->o_label._labelName)
#define srfd_set_objectLabel(_object, _labelName, _label)  \
	do { \
		struct SRFD_lbl old_label = _object->o_label._labelName; \
		if (srfd_label_eq(old_label, _label)) break; \
		(_object->o_label._labelName = _label); \
		srfd_logDowngrade("object ino %lu " #_labelName " is changed from %d to %d", _object->o_ino, SRFD_label2int(old_label), SRFD_label2int(_label)); \
	} while (0)

#define srfd_get_objectLabelV(_object, _labelName) (srfd_get_objectLabel(_object, _labelName).level)


void srfd_object_markAsInvalid(struct SRFD_object *object);





struct SRFD_handle {
	atomic_t		h_rc;
	SRFD_HANDLE_TYPE		h_type;

	struct list_head	h_list;

	struct list_head	h_sg_list;	/* subgroup's handle list */
	struct list_head	h_o_list;	/* object's handle list */

	struct SRFD_subgroup	*h_sg;
	struct SRFD_object	*h_o;

#define h_isValid		h_o->o_isValid

	struct SRFD_handle_lbl 	h_label;

	unsigned int 		h_canHandleErrors;
	unsigned int		h_isFlowMediated;
	unsigned int		h_isDelayed;
	int			h_fd;

	struct rcu_head		h_rcu;

//	int			h_isMarkedAsInvalid;
//	int			h_isPendingForDeletion;
};


struct SRFD_handle *srfd_get_handle(struct SRFD_handle *handle);
void srfd_handle_free(struct SRFD_handle *handle);
void srfd_put_handle(struct SRFD_handle *handle);
struct SRFD_handle *srfd_handle_alloc(void);
int srfd_handle_matchWithFile(struct SRFD_handle *handle, struct file *file);
struct SRFD_handle *srfd_handle_dupWithNewSubgroup(struct SRFD_handle *oldHandle, struct SRFD_subgroup *newSubgroup);
int srfd_handle_isNoLongerValid(struct SRFD_handle *handle);
//void srfd_handle_markAsInvalid(struct SRFD_handle *handle);


#define srfd_get_handleLabel(_handle, _labelName) (_handle->h_label._labelName)
#define srfd_set_handleLabel(_handle, _labelName, _label) (_handle->h_label._labelName = _label)

#define srfd_get_handleLabelV(_handle, _labelName) (srfd_get_handleLabel(_handle, _labelName).level)




/* handle and object */
int srfd_addHandle2Object(struct SRFD_handle *handle, struct SRFD_object *object);
void srfd_removeHandleFromObject(struct SRFD_handle *handle);


/* handle and subgroup */
int srfd_addHandle2Subgroup(struct SRFD_handle *handle, struct SRFD_subgroup *subgroup);
int srfd_removeHandleFromSubgroup(struct SRFD_handle *handle);


/* current_min_lbl constraint propagation */
struct SRFD_lbl srfd_computeHandle_new_current_min_lbl(struct SRFD_handle *handle, struct SRFD_lbl upstream_current_min_lbl);
int srfd_current_min_lbl_canPropagateFromHandle(struct SRFD_handle *handle, struct SRFD_lbl upstream_current_min_lbl);

#define srfd_noConstraintOnHandle_current_min_lbl(_handle) (_handle->h_canHandleErrors && _handle->h_isFlowMediated)


#define srfd_handle_resetNewMinToCurMin(_handle) srfd_set_handleLabel(_handle, new_current_min_lbl, srfd_get_handleLabel(_handle, current_min_lbl))

#define srfd_object_resetNewMinToCurMin(_object) srfd_set_objectLabel(_object, new_current_min_lbl, srfd_get_objectLabel(_object, current_min_lbl))

#define srfd_subgroup_resetNewMinToCurMin(_subgroup) srfd_set_subgroupLabel(_subgroup, new_current_min_lbl, srfd_get_subgroupLabel(_subgroup, current_min_lbl))

int srfd_current_min_lbl_canPropagateFromObject(struct SRFD_object *object, struct SRFD_lbl upstream_current_min_lbl);
int srfd_current_min_lbl_canPropagateFromSubgroup(struct SRFD_subgroup *subgroup, struct SRFD_lbl upstream_current_min_lbl);

int srfd_new_current_min_lbl_canPropagateDownstreamFromHandle(struct SRFD_handle *handle);





#define srfd_handle_current_min_lbl_propagated(_handle) (srfd_label_eq(srfd_get_handleLabel(_handle, current_min_lbl), srfd_get_handleLabel(_handle, new_current_min_lbl)))
#define srfd_object_current_min_lbl_propagated(_object) (srfd_label_eq(srfd_get_objectLabel(_object, current_min_lbl), srfd_get_objectLabel(_object, new_current_min_lbl)))
#define srfd_subgroup_current_min_lbl_propagated(_subgroup) (srfd_label_eq(srfd_get_subgroupLabel(_subgroup, current_min_lbl), srfd_get_subgroupLabel(_subgroup, new_current_min_lbl)))

#define srfd_handle_update_current_min_lbl(_handle) \
        do { \
                srfd_set_handleLabel(_handle, current_min_lbl, \
                        srfd_get_handleLabel(_handle, new_current_min_lbl)); \
        } while (0)

#define srfd_object_update_current_min_lbl(_object) \
	do { \
		srfd_set_objectLabel(_object, current_min_lbl, \
			srfd_get_objectLabel(_object, new_current_min_lbl)); \
        } while (0)

#define srfd_subgroup_update_current_min_lbl(_subgroup) \
	do { \
		srfd_set_subgroupLabel(_subgroup, current_min_lbl, \
			srfd_get_subgroupLabel(_subgroup, new_current_min_lbl)); \
	} while (0)


#define srfd_handle_increase_current_min_lbl(_handle) \
        do { \
                srfd_assert1(srfd_label_ge(srfd_get_handleLabel(_handle, new_current_min_lbl), \
                        srfd_get_handleLabel(_handle, current_min_lbl))); \
                srfd_set_handleLabel(_handle, current_min_lbl, \
                        srfd_get_handleLabel(_handle, new_current_min_lbl)); \
        } while (0)

#define srfd_object_increase_current_min_lbl(_object) \
	do { \
		srfd_assert1(srfd_label_ge(srfd_get_objectLabel(_object, new_current_min_lbl), \
			srfd_get_objectLabel(_object, current_min_lbl))); \
		srfd_set_objectLabel(_object, current_min_lbl, \
			srfd_get_objectLabel(_object, new_current_min_lbl)); \
        } while (0)

#define srfd_subgroup_increase_current_min_lbl(_subgroup) \
	do { \
		srfd_assert1(srfd_label_ge(srfd_get_subgroupLabel(_subgroup, new_current_min_lbl), \
			srfd_get_subgroupLabel(_subgroup, current_min_lbl))); \
		srfd_set_subgroupLabel(_subgroup, current_min_lbl, \
			srfd_get_subgroupLabel(_subgroup, new_current_min_lbl)); \
	} while (0)


void srfd_current_min_lbl_updateWithNewFromHandle(struct SRFD_handle *handle);
void srfd_current_min_lbl_updateWithNewFromObject(struct SRFD_object *object);
void srfd_current_min_lbl_updateWithNewFromSubgroup(struct SRFD_subgroup *subgroup);

int srfd_current_min_lbl_subgroupCanDecrease2Target(struct SRFD_subgroup *subgroup, struct SRFD_lbl target_min_lbl);
int srfd_current_min_lbl_objectCanDecrease2Target(struct SRFD_object *object, struct SRFD_lbl target_min_lbl);
int srfd_current_min_lbl_handleCanDecrease2Target(struct SRFD_handle *handle, struct SRFD_lbl target_min_lbl);
int srfd_current_min_lbl_canDecreaseDownstream(struct SRFD_handle *handle, struct SRFD_lbl target_min_lbl);

/* Increase min_lbl fails when current < target */
int srfd_current_min_lbl_subgroupCanIncrease2Target(struct SRFD_subgroup *subgroup, struct SRFD_lbl target_min_lbl);
int srfd_current_min_lbl_objectCanIncrease2Target(struct SRFD_object *object, struct SRFD_lbl target_min_lbl);
int srfd_current_min_lbl_handleCanIncrease2Target(struct SRFD_handle *handle, struct SRFD_lbl target_min_lbl);
int srfd_current_min_lbl_canIncreaseDownstream(struct SRFD_handle *handle, struct SRFD_lbl target_min_lbl);



void srfd_current_min_lbl_propagateFromHandle(struct SRFD_handle *handle);
void srfd_current_min_lbl_propagateFromObject(struct SRFD_object *object);
void srfd_current_min_lbl_propagateFromSubgroup(struct SRFD_subgroup *subgroup);

void srfd_current_min_lbl_subgroupRecompute(struct SRFD_subgroup *subgroup);
void srfd_current_min_lbl_objectRecompute(struct SRFD_object *object);
void srfd_current_min_lbl_handleRecompute(struct SRFD_handle *handle);

/* srfd_current_lbl */

void srfd_current_lbl_propagateFromSubgroup(struct SRFD_subgroup *subgroup);
void srfd_current_lbl_propagateFromObject(struct SRFD_object *object);
void srfd_current_lbl_propagateFromHandle(struct SRFD_handle *handle);



void srfd_subgroup_setFilesStruct(struct SRFD_subgroup *subgroup, struct files_struct *files);
void srfd_printStatistics(void);


/* subpid and subgroup */
int __srfd_addSubpid2Subgroup(struct SRFD_subpid *subpid, struct SRFD_subgroup *subgroup);
int srfd_removeSubpidFromSubgroup(struct SRFD_subpid *subpid);
int srfd_subgroup_isValid(struct SRFD_subgroup *subgroup);
int srfd_subpid_alreadyInSubgroup(struct SRFD_subpid *subpid2find, struct SRFD_subgroup *subgroup);
#define srfd_addSubpid2Subgroup(_subpid, _subgroup) \
	do { \
		if (srfd_subpid_alreadyInSubgroup(_subpid, _subgroup)) \
			break; \
		__srfd_addSubpid2Subgroup(_subpid, _subgroup); \
		_subgroup->sg_initState = SRFD_SUBGROUP_INIT_STATE_COMPLETE; \
	} while (0)
char *srfd_subgroup_getProcessesList(struct SRFD_subgroup *subgroup);
#define srfd_subgroup_freeProcessesList(_buf) free_page((long unsigned int)(_buf))




#define srfd_do_foreach_container_all_handleType(_container, _handle, _container_handle_list_lock, _container_handle_list, _handle_container_list) { \
/*	unsigned long flags; */\
/*	spin_lock_irqsave(&_container->_container_handle_list_lock, flags);*/ \
	list_for_each_entry(_handle, &_container->_container_handle_list, _handle_container_list) {



#define srfd_do_foreach_container_valid_handleType(_container, _handle, _container_handle_list_lock, _container_handle_list, _handle_container_list) { \
/*	unsigned long flags; */\
/*	spin_lock_irqsave(&_container->_container_handle_list_lock, flags);*/ \
	list_for_each_entry(_handle, &_container->_container_handle_list, _handle_container_list) { \
		{ if (!_handle->h_isValid || srfd_handle_isNoLongerValid(_handle)) \
			continue; }


#define srfd_do_foreach_container_valid_handleType_rm_invalid(_container, _handle, _container_handle_list_lock, _container_handle_list, _handle_container_list) { \
/*	unsigned long flags; */\
/*	spin_lock_irqsave(&_container->_container_handle_list_lock, flags);*/ \
	list_for_each_entry(_handle, &_container->_container_handle_list, _handle_container_list) { \
		{ if (!_handle->h_isValid || srfd_handle_isNoLongerValid(_handle)) { \
			srfd_removeHandleFromObject(handle); \
			srfd_removeHandleFromSubgroup(handle); \
			srfd_put_handle(handle); \
			continue; \
		}}






#define srfd_beforebreak_foreach_container_handleType(_container, _container_handle_list_lock) \
/*	spin_unlock_irqrestore(&_container->_container_handle_list_lock, flags)*/
	
#define srfd_while_foreach_container_handleType(_container, _container_handle_list_lock) \
	} /*spin_unlock_irqrestore(&_container->_container_handle_list_lock, flags);*/ \
	}

/* Object, Handle */
#define srfd_do_foreach_object_all_rh(_object, _handle) \
	srfd_do_foreach_container_all_handleType(_object, _handle, o_rh_list_lock, o_rh_list, h_o_list)

#define srfd_do_foreach_object_valid_rh(_object, _handle) \
	srfd_do_foreach_container_valid_handleType(_object, _handle, o_rh_list_lock, o_rh_list, h_o_list)

#define srfd_beforebreak_foreach_object_rh(_object) \
	srfd_beforebreak_foreach_container_handleType(_object, o_rh_list_lock)

#define srfd_while_foreach_object_rh(_object) \
	srfd_while_foreach_container_handleType(_object, o_rh_list_lock)


#define srfd_do_foreach_object_all_wh(_object, _handle) \
	srfd_do_foreach_container_all_handleType(_object, _handle, o_wh_list_lock, o_wh_list, h_o_list)

#define srfd_do_foreach_object_valid_wh(_object, _handle) \
	srfd_do_foreach_container_valid_handleType(_object, _handle, o_wh_list_lock, o_wh_list, h_o_list)

#define srfd_beforebreak_foreach_object_wh(_object) \
	srfd_beforebreak_foreach_container_handleType(_object, o_wh_list_lock)

#define srfd_while_foreach_object_wh(_object) \
	srfd_while_foreach_container_handleType(_object, o_wh_list_lock)


/* Subject, Handle */
#define srfd_do_foreach_subgroup_all_rh(_subgroup, _handle) \
	srfd_do_foreach_container_all_handleType(_subgroup, _handle, sg_rh_list_lock, sg_rh_list, h_sg_list)

#define srfd_do_foreach_subgroup_valid_rh(_subgroup, _handle) \
	srfd_do_foreach_container_valid_handleType(_subgroup, _handle, sg_rh_list_lock, sg_rh_list, h_sg_list)

#define srfd_do_foreach_subgroup_valid_rh_rm_invalid(_subgroup, _handle) \
	srfd_do_foreach_container_valid_handleType_rm_invalid(_subgroup, _handle, sg_rh_list_lock, sg_rh_list, h_sg_list)

#define srfd_do_foreach_subgroup_valid_wh_rm_invalid(_subgroup, _handle) \
	srfd_do_foreach_container_valid_handleType_rm_invalid(_subgroup, _handle, sg_wh_list_lock, sg_wh_list, h_sg_list)




#define srfd_beforebreak_foreach_subgroup_rh(_subgroup) \
	srfd_beforebreak_foreach_container_handleType(_subgroup, sg_rh_list_lock)

#define srfd_while_foreach_subgroup_rh(_subgroup) \
	srfd_while_foreach_container_handleType(_subgroup, sg_rh_list_lock)




#define srfd_do_foreach_subgroup_all_wh(_subgroup, _handle) \
	srfd_do_foreach_container_all_handleType(_subgroup, _handle, sg_wh_list_lock, sg_wh_list, h_sg_list)

#define srfd_do_foreach_subgroup_valid_wh(_subgroup, _handle) \
	srfd_do_foreach_container_valid_handleType(_subgroup, _handle, sg_wh_list_lock, sg_wh_list, h_sg_list)

#define srfd_beforebreak_foreach_subgroup_wh(_subgroup) \
	srfd_beforebreak_foreach_container_handleType(_subgroup, sg_wh_list_lock)

#define srfd_while_foreach_subgroup_wh(_subgroup) \
	srfd_while_foreach_container_handleType(_subgroup, sg_wh_list_lock)






#define SRFD_GET_READ_HANDLE_IN_SUBGROUP_BY_INODE(_subgroup, _inode) SRFD_GET_HANDLE_IN_SUBGROUP_BY_INODE(_subgroup, _inode, rh)

#define SRFD_GET_WRITE_HANDLE_IN_SUBGROUP_BY_INODE(_subgroup, _inode) SRFD_GET_HANDLE_IN_SUBGROUP_BY_INODE(_subgroup, _inode, wh)


#define SRFD_GET_HANDLE_IN_SUBGROUP_BY_INODE(_subgroup, _inode, _handle_type) \
	({ \
		struct SRFD_handle *handle = NULL, *ret_handle = NULL; \
		srfd_do_foreach_subgroup_valid_ ##_handle_type(_subgroup, handle) { \
			if (handle->h_o->o_inode == _inode) { \
				ret_handle = handle; \
				srfd_beforebreak_foreach_subgroup_ ##_handle_type(_subgroup); \
				break; \
			} \
		} srfd_while_foreach_subgroup_ ##_handle_type(_subgroup); \
		(srfd_get_handle(ret_handle)); \
	})

#define SRFD_PUT_HANDLE_IN_SUBGROUP_BY_INODE(_handle) \
	srfd_put_handle(_handle);


#define SRFD_GET_READ_HANDLE_IN_OBJECT_BY_SUBGROUP(_object, _subgroup) SRFD_GET_HANDLE_IN_OBJECT_BY_SUBGROUP(_object, _subgroup, rh)

#define SRFD_GET_WRITE_HANDLE_IN_OBJECT_BY_SUBGROUP(_object, _subgroup) SRFD_GET_HANDLE_IN_OBJECT_BY_SUBGROUP(_object, _subgroup, wh)


#define SRFD_GET_HANDLE_IN_OBJECT_BY_SUBGROUP(_object, _subgroup, _handle_type) \
	({ \
		struct SRFD_handle *handle = NULL, *ret_handle = NULL; \
		srfd_do_foreach_object_valid_ ##_handle_type(_object, handle) { \
			if (handle->h_sg == _subgroup) { \
				ret_handle = handle; \
				srfd_beforebreak_foreach_object_ ##_handle_type(_object); \
				break; \
			} \
		} srfd_while_foreach_object_ ##_handle_type(_object); \
		(srfd_get_handle(ret_handle)); \
	})


#define SRFD_PUT_HANDLE_IN_OBJECT_BY_SUBGROUP(_handle) \
	srfd_put_handle(_handle);








#define SRFD_GET_CURRENT_SUBGROUP() \
	({ \
		struct SRFD_subgroup *subgroup = NULL; \
		struct SRFD_subpid *subpid; \
		struct SRFD_credsec *credsec = srfd_credsec_findAndGetByCred(current_cred()); \
		if (srfd_assert1r(credsec != NULL)) \
			goto __out_sg; \
		subgroup = srfd_get_subgroup(credsec->cs_sg); \
		srfd_put_credsec(credsec); \
		if (srfd_assert1r(subgroup != NULL)) \
			goto __out_sg; \
		subpid = srfd_subpid_findAndGet(task_pid(current)); \
		if (subpid == NULL) { \
			subpid = srfd_get_subpid(srfd_subpid_allocAndGet(task_pid(current))); \
			if (srfd_assert1r(subpid != NULL)) \
				goto __out_sg; \
			srfd_addSubpid2Subgroup(subpid, subgroup); \
		} \
		if (credsec->cs_parent_cs != NULL) { \
			struct SRFD_handle *newHandle, *handle; \
			struct SRFD_subgroup *old_subgroup = srfd_get_subgroup(credsec->cs_parent_cs->cs_sg); \
			if (srfd_assertr(old_subgroup != NULL, "parent credsec subgroup is NULL")) \
				goto __put_parent_credsec; \
			if (srfd_assertr(srfd_subgroup_isValid(old_subgroup), "parent credsec subgroup is invalid!")) \
				goto __put_parent_subgroup; \
			srfd_do_foreach_subgroup_valid_rh(old_subgroup, handle) { \
				newHandle = srfd_handle_dupWithNewSubgroup(handle, subgroup); \
			} srfd_while_foreach_subgroup_rh(old_subgroup); \
			srfd_do_foreach_subgroup_valid_wh(old_subgroup, handle) { \
				newHandle = srfd_handle_dupWithNewSubgroup(handle, subgroup); \
			} srfd_while_foreach_subgroup_wh(old_subgroup); \
__put_parent_subgroup: \
			srfd_put_subgroup(old_subgroup); \
__put_parent_credsec: \
			srfd_put_credsec(credsec->cs_parent_cs); \
			credsec->cs_parent_cs = NULL; \
		} \
		srfd_put_subpid(subpid); \
__out_sg:  \
		(subgroup); \
	})


#define SRFD_PUT_CURRENT_SUBGROUP(_subgroup) srfd_put_subgroup(_subgroup)


#define SRFD_GET_OBJECT_BY_INODE(inode) \
	({ \
		struct SRFD_object *object = srfd_object_findAndGetByInode(inode); \
		if (object == NULL) { \
			object = srfd_get_object(srfd_object_allocAndGet(inode)); \
			srfd_assert1(object != NULL); \
		} \
		(object); \
	})

/*			if (srfd_assert1r(object != NULL)) \
				goto __out_o; \
		} \
__out_o: \
		(object); \
	})
*/
#define SRFD_PUT_OBJECT(_object) srfd_put_object(_object)

















#endif

