/* * Self-Revocation Free Downgrading - A kernel module that applies information flow
 * control to protect system integrity without self-revocation
 * Copyright (C) 2014 Secure Systems Laboratory, Stony Brook University
 *
 * This file is part of Self-Revocation Free Downgrading.
 *
 * Self-Revocation Free Downgrading is free software: you can redistribute it
 * and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Self-Revocation Free Downgrading is distributed in the hope that it will be
 * useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Self-Revocation Free Downgrading.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

#include "srfd_security.h"
#include "srfd_security_internal.h"
#include "srfd_internal.h"

static DEFINE_SPINLOCK(srfd_subgroup_list_lock);

struct list_head srfd_subgroup_list[SRFD_MAX_SUBGROUP_HASH];
atomic_t srfd_subgroup_activeCount = ATOMIC_INIT(0);


#define subgroup2hash(subgroupPtr) \
	&srfd_subgroup_list[hash_ptr((void *)subgroupPtr, SRFD_SUBGROUP_HASH_BITS)]

struct SRFD_subject_lbl srfd_security_subject_default_label = {
	.can_handle_read_errors = 0,
	.can_handle_write_errors = 0,
	.current_lbl = {7},
	.min_lbl = {1},
	.current_min_lbl = {1},

	.new_current_min_lbl = {1},

	.invul_lbl = {7},
	.discr_invul_lbl = {7},
	.discr_input_invul = {7},
	.discr_output_invul = {7}

};




void __srfd_subgroup_add2List(struct SRFD_subgroup *ptr) {
	unsigned long flags;
	struct list_head *list = subgroup2hash(ptr);

	spin_lock_irqsave(&srfd_subgroup_list_lock, flags);
	list_add_rcu(&ptr->sg_list, list);
	spin_unlock_irqrestore(&srfd_subgroup_list_lock, flags);
	atomic_inc(&srfd_subgroup_activeCount);
}

static void __srfd_subgroup_rcuFree(struct rcu_head *rcu) {
	struct SRFD_subgroup *ptr = container_of(rcu, typeof(*ptr), sg_rcu);
	kfree(ptr);
	atomic_dec(&srfd_subgroup_activeCount);
}

void srfd_subgroup_free(struct SRFD_subgroup *subgroup) {
	
	unsigned long flags;

	if (atomic_read(&subgroup->sg_rc) != 0)
		srfd_info("Freeing subject group with usage count: %d != 0 %p", atomic_read(&subgroup->sg_rc), subgroup);

	spin_lock_irqsave(&srfd_subgroup_list_lock, flags);
	list_del_rcu(&subgroup->sg_list);
	spin_unlock_irqrestore(&srfd_subgroup_list_lock, flags);
	call_rcu(&subgroup->sg_rcu, __srfd_subgroup_rcuFree);
}

struct SRFD_subgroup *srfd_get_subgroup(struct SRFD_subgroup *subgroup) {
	if (subgroup != NULL)
		atomic_inc(&subgroup->sg_rc);
	return subgroup;
}

void srfd_put_subgroup(struct SRFD_subgroup *subgroup) {
	if (atomic_dec_and_test(&subgroup->sg_rc))
		srfd_subgroup_free(subgroup);
}


struct SRFD_subgroup *srfd_subgroup_alloc(void) {

	struct SRFD_subgroup *subgroup = kzalloc(sizeof(*subgroup), GFP_ATOMIC);

	if (srfd_assert1r(subgroup != NULL))
		return NULL;

        atomic_set(&subgroup->sg_rc, 0);
        atomic_set(&subgroup->sg_cred_rc, 0);

	subgroup->sg_files = NULL;
	__srfd_subgroup_add2List(subgroup);

	INIT_LIST_HEAD(&subgroup->sg_rh_list);
	INIT_LIST_HEAD(&subgroup->sg_wh_list);
	INIT_LIST_HEAD(&subgroup->sg_sp_list);

	subgroup->sg_rh_list_lock = __SPIN_LOCK_UNLOCKED(subgroup->sg_rh_list_lock);
	subgroup->sg_wh_list_lock = __SPIN_LOCK_UNLOCKED(subgroup->sg_wh_list_lock);
	subgroup->sg_sp_list_lock = __RW_LOCK_UNLOCKED(subgroup->sg_sp_list_lock);

	subgroup->sg_isNewGroup = 0;
	subgroup->sg_label = srfd_security_subject_default_label;

	subgroup->sg_initState = SRFD_SUBGROUP_INIT_STATE_COMPLETE;

	return subgroup;

}



struct SRFD_subgroup *srfd_subgroup_dup(struct SRFD_subgroup *origSubgroup) {

	struct SRFD_subgroup *newSubgroup = NULL;
	srfd_info("dup subgroup: original subgroup is %p", origSubgroup);

	if (srfd_assert1r(origSubgroup != NULL))
		goto out;

	newSubgroup = srfd_subgroup_alloc();
	if (srfd_assert1r(newSubgroup != NULL))
		goto out;

	newSubgroup->sg_label = origSubgroup->sg_label;
out:
	return newSubgroup;

}



