/* * Self-Revocation Free Downgrading - A kernel module that applies information flow
 * control to protect system integrity without self-revocation
 * Copyright (C) 2014 Secure Systems Laboratory, Stony Brook University
 *
 * This file is part of Self-Revocation Free Downgrading.
 *
 * Self-Revocation Free Downgrading is free software: you can redistribute it
 * and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Self-Revocation Free Downgrading is distributed in the hope that it will be
 * useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Self-Revocation Free Downgrading.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

#include "srfd_security_internal.h"
#include "srfd_debug.h"

static DEFINE_SPINLOCK(srfd_subpid_list_lock);

struct list_head srfd_subpid_list[SRFD_MAX_SUBPID_HASH];
atomic_t srfd_subpid_activeCount = ATOMIC_INIT(0);


#define pid2hash(pid) \
	&srfd_subpid_list[hash_ptr((pid), SRFD_SUBPID_HASH_BITS)]


void __srfd_subpid_add2List(struct SRFD_subpid *subpid) {
	unsigned long flags;
	struct list_head *list = pid2hash(subpid->sp_pid);

	spin_lock_irqsave(&srfd_subpid_list_lock, flags);
	list_add_rcu(&subpid->sp_list, list);
	spin_unlock_irqrestore(&srfd_subpid_list_lock, flags);

	atomic_inc(&srfd_subpid_activeCount);
}


struct SRFD_subpid *srfd_subpid_alloc(struct pid *pid) {
	struct SRFD_subpid *subpid = kzalloc(sizeof(*subpid), GFP_ATOMIC);

	if (srfd_assertr(subpid != NULL, "Cannot allocate memory for task"))
		return NULL;
	
	subpid->sp_pid = get_pid(pid);
	atomic_set(&subpid->sp_rc, 0);
	subpid->sp_sg = NULL;

	__srfd_subpid_add2List(subpid);

	return subpid;
}


struct SRFD_subpid *srfd_subpid_findByPid(struct pid *pid) {

	struct list_head *list = pid2hash(pid);
	struct SRFD_subpid *subpid;

	while (!list->next);
	rcu_read_lock();
	list_for_each_entry(subpid, list, sp_list) {
		if (subpid->sp_pid != pid)
			continue;
		rcu_read_unlock();
		return subpid;
	}
	rcu_read_unlock();
	return NULL;
}



static void __srfd_subpid_rcuFree(struct rcu_head *rcu) {
	struct SRFD_subpid *subpid = container_of(rcu, typeof(*subpid), sp_rcu);
	put_pid(subpid->sp_pid);

	kfree(subpid);
	atomic_dec(&srfd_subpid_activeCount);	
}



void srfd_subpid_free(struct SRFD_subpid *subpid) {
	unsigned long flags;

	srfd_assert(atomic_read(&subpid->sp_rc) == 0, "freeing subpid with rc %d", atomic_read(&subpid->sp_rc));

	spin_lock_irqsave(&srfd_subpid_list_lock, flags);
	list_del_rcu(&subpid->sp_list);
	spin_unlock_irqrestore(&srfd_subpid_list_lock, flags);

	call_rcu(&subpid->sp_rcu, __srfd_subpid_rcuFree);

}

int srfd_subpid_isNoLongerValid(struct SRFD_subpid *subpid) {
	if (srfd_assert1r(subpid != NULL))
		return 1;
	return (pid_task(subpid->sp_pid, PIDTYPE_PID) == NULL);
}

struct SRFD_subpid *srfd_get_subpid(struct SRFD_subpid *subpid) {
	if (subpid != NULL)
		atomic_inc(&subpid->sp_rc);
	return subpid;
}

void srfd_put_subpid(struct SRFD_subpid *subpid) {
	if (atomic_dec_and_test(&subpid->sp_rc))
		srfd_subpid_free(subpid);
}

