/* * Self-Revocation Free Downgrading - A kernel module that applies information flow
 * control to protect system integrity without self-revocation
 * Copyright (C) 2014 Secure Systems Laboratory, Stony Brook University
 *
 * This file is part of Self-Revocation Free Downgrading.
 *
 * Self-Revocation Free Downgrading is free software: you can redistribute it
 * and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Self-Revocation Free Downgrading is distributed in the hope that it will be
 * useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Self-Revocation Free Downgrading.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

#include "srfd_security.h"
#include "srfd_debug.h"
#include <linux/mm.h>

int __srfd_addSubpid2Subgroup(struct SRFD_subpid *subpid, struct SRFD_subgroup *subgroup) {
	unsigned long flags;

	write_lock_irqsave(&subgroup->sg_sp_list_lock, flags);
	list_add(&subpid->sp_sg_list, &subgroup->sg_sp_list);
	write_unlock_irqrestore(&subgroup->sg_sp_list_lock, flags);

	subpid->sp_sg = srfd_get_subgroup(subgroup);

	return 0;
}


int srfd_removeSubpidFromSubgroup(struct SRFD_subpid *subpid) {

	unsigned long flags;
	struct SRFD_subgroup *subgroup = subpid->sp_sg;

	if (srfd_assertr(subgroup != NULL, "removing subpid %p from NULL subgroup", subpid))
		goto out;

	write_lock_irqsave(&subgroup->sg_sp_list_lock, flags);
	list_del(&subpid->sp_sg_list);
	write_unlock_irqrestore(&subgroup->sg_sp_list_lock, flags);

	subpid->sp_sg = NULL;
	srfd_put_subgroup(subgroup);
out:
	return 0;
}


int srfd_subgroup_isValid(struct SRFD_subgroup *subgroup) {
	struct SRFD_subpid *subpid, *subpid_n;
	int isValid = 0;

	list_for_each_entry_safe(subpid, subpid_n, &subgroup->sg_sp_list, sp_sg_list) {
		if (srfd_subpid_isNoLongerValid(subpid)) {
			srfd_removeSubpidFromSubgroup(subpid);
			srfd_put_subpid(subpid);
			continue;
		} else
			return 1;
	}

	return isValid;
}


int srfd_subpid_alreadyInSubgroup(struct SRFD_subpid *subpid2find, struct SRFD_subgroup *subgroup) {
	struct SRFD_subpid *subpid, *subpid_n;

	list_for_each_entry_safe(subpid, subpid_n, &subgroup->sg_sp_list, sp_sg_list) {
		if (srfd_subpid_isNoLongerValid(subpid)) {
			srfd_removeSubpidFromSubgroup(subpid);
			continue;
		} else if (subpid2find == subpid)
			return 1;
	}
	return 0;
}


char *srfd_subgroup_getProcessesList(struct SRFD_subgroup *subgroup) {

#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 32)
	struct SRFD_subpid *subpid, *subpid_n;
	int printed = 0;
	struct task_struct *task;
	char *result = NULL; 

	if (!srfd_get_subgroup(subgroup))
		return NULL;

	result = (char *)__get_free_page(GFP_KERNEL);

	if (result && !IS_ERR(result)) {
		result[0] = 0;
		list_for_each_entry_safe(subpid, subpid_n, &subgroup->sg_sp_list, sp_sg_list) {
			if (srfd_subpid_isNoLongerValid(subpid)) {
				continue;
			}
			task = get_pid_task(subpid->sp_pid, PIDTYPE_PID);
			if (task) {
				printed += snprintf(result + printed, PAGE_SIZE-printed, "%s (%d), ", task->comm, task->pid);
				put_task_struct(task);
			}
		}
		if (printed > 2)
			result[printed-2] = 0;

	} else
		result = NULL;

	srfd_put_subgroup(subgroup);

	return result;
#else
	return NULL;
#endif
}






