/* * Self-Revocation Free Downgrading - A kernel module that applies information flow
 * control to protect system integrity without self-revocation
 * Copyright (C) 2014 Secure Systems Laboratory, Stony Brook University
 *
 * This file is part of Self-Revocation Free Downgrading.
 *
 * Self-Revocation Free Downgrading is free software: you can redistribute it
 * and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Self-Revocation Free Downgrading is distributed in the hope that it will be
 * useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Self-Revocation Free Downgrading.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

#include <sys/socket.h>
#include <sys/un.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stddef.h>
#include <errno.h>

char *socket_path_high = "hiddenHigh";
char *socket_path_low = "hiddenLow";

const char *LOWI_FILE = "../testfiles/lowI";
const char *HIGHI_FILE = "../testfiles/highI";

FILE *highI = NULL, *lowI = NULL;
int peerSocketHigh = -1;
int peerSocketLow = -1;


char buf[100];

#define PPI_LOGG(format, args...) printf(format "\n", ##args)

#define PPI_OK PPI_LOGG
#define PPI_ERROR PPI_LOGG
#define PPI_NOT_EXPECTED PPI_LOGG

#define PPI_LOG(_level, _format, _args...) PPI_ ##_level( "[" #_level "] " _format, ##_args)


#define PPI_EXPECTED(_x) do {\
		if (!(_x)) { \
			PPI_LOG(NOT_EXPECTED, "Assert failed: " #_x); \
			PPI_ERROR("Aborting current test: " currentTestName); \
			goto nextTest; \
		} \
	} while (0)


FILE *openHighI(char *_mode) {
	highI = fopen(HIGHI_FILE, _mode);
	return highI;
}

FILE *openLowI(char *_mode) {
	lowI = fopen(LOWI_FILE, _mode);
	return lowI;
}



int connect2Low() {
	int len;
	char bindPath[100];
	struct sockaddr_un addr;

	if (peerSocketLow == -1) {
		peerSocketLow = socket(AF_UNIX, SOCK_STREAM, 0);

		bzero(&addr, sizeof(addr));
		addr.sun_family = AF_UNIX;
		bindPath[0] = 0;
		len = sprintf(bindPath+1, "%s", socket_path_low) + 1;
		memcpy(addr.sun_path, bindPath, len);
		len += offsetof(struct sockaddr_un, sun_path);

		if (connect(peerSocketLow, (struct sockaddr*)&addr, len) == -1) {
			peerSocketLow = -1;
		}
	}
	return peerSocketLow;
}

int readFromLowPeer() {
	int rv = -1;
	char buf[10];
	if (peerSocketLow != -1) {
		rv = recv(peerSocketLow, buf, 1, 0);
	}
	return rv;
}




int connect2High() {

	int len;
	char bindPath[100];
	struct sockaddr_un addr;

	if (peerSocketHigh == -1) {
		peerSocketHigh = socket(AF_UNIX, SOCK_STREAM, 0);

		bzero(&addr, sizeof(addr));
		addr.sun_family = AF_UNIX;
		bindPath[0] = 0;
		len = sprintf(bindPath+1, "%s", socket_path_high) + 1;
		memcpy(addr.sun_path, bindPath, len);
		len += offsetof(struct sockaddr_un, sun_path);

		if (connect(peerSocketHigh, (struct sockaddr*)&addr, len) == -1) {
			peerSocketHigh = -1;
		}
	}
	return peerSocketHigh;
}

void flushEnv() {
	if (highI) {
		fclose(highI);
		highI = NULL;
	}
	if (lowI) {
		fclose(lowI);
		lowI = NULL;
	}
	if (peerSocketHigh != -1) {
		close(peerSocketHigh);
		peerSocketHigh = -1;
	}
	if (peerSocketLow != -1) {
		close(peerSocketLow);
		peerSocketLow = -1;
	}
}






int main(int argc, char *argv[]) {


	if (connect2High() == -1) {
		PPI_LOG(NOT_EXPECTED, "Cannot connect to peer process without any opened files");
		return -1;
	}
	flushEnv();


#define nextTest test2
#define currentTestName "Test 1: Can connect to peer HighI process when nothing opened"
	PPI_EXPECTED(connect2High());
	PPI_LOG(OK, currentTestName);

test2:

#undef nextTest
#define nextTest test3
#undef currentTestName
#define currentTestName "Test 2: Cannot connect to HighI process when lowI file is opened in read mode"
	/* Test 2:
		With high integrity file opened in read mode, low integrity file can be opened
	*/
	flushEnv();
	PPI_EXPECTED(openLowI("r+"));
	PPI_EXPECTED(connect2High() == -1);
	PPI_LOG(OK, currentTestName);

test3:

#undef nextTest
#define nextTest test4
#undef currentTestName
#define currentTestName "Test 3: Can connect to HighI process when highI file is opened in read/write mode"
	/* Test 3:
		With high integrity file opened in read/write mode, can connect to peer
	*/
	flushEnv();
	PPI_EXPECTED(openHighI("r+"));
	PPI_EXPECTED(connect2High() != -1);
	PPI_LOG(OK, currentTestName);

test4:

#undef nextTest
#define nextTest test5
#undef currentTestName
#define currentTestName "Test 4: Cannot open lowI for reading after connected to HighI process"
	flushEnv();
	PPI_EXPECTED(connect2High() != -1);
	PPI_EXPECTED(openLowI("r+") == NULL);
	PPI_LOG(OK, currentTestName);


test5:

#undef nextTest
#define nextTest test6
#undef currentTestName
#define currentTestName "Test 5: Can open highI for reading after connected to HighI process"
	flushEnv();
	PPI_EXPECTED(connect2High() != -1);
	PPI_EXPECTED(openHighI("r+") != NULL);
	PPI_LOG(OK, currentTestName);


test6:

#undef nextTest
#define nextTest test7
#undef currentTestName
#define currentTestName "Test 6: Can open lowI for write-only after connected to HighI process"
	flushEnv();
	PPI_EXPECTED(connect2High() != -1);
	PPI_EXPECTED(openLowI("w") != NULL);
	PPI_LOG(OK, currentTestName);


test7:

#undef nextTest
#define nextTest test8
#undef currentTestName
#define currentTestName "Test 7: Cannot connect to LowI process when opened HighI for writing"
	flushEnv();
	PPI_EXPECTED(openHighI("r+") != NULL);
	PPI_EXPECTED(connect2Low() == -1);
	PPI_LOG(OK, currentTestName);

test8:

#undef nextTest
#define nextTest test9
#undef currentTestName
#define currentTestName "Test 8: Can connect to LowI process when opened HighI for reading. (Will downgrade process)"
	flushEnv();
	PPI_EXPECTED(openHighI("r") != NULL);
	PPI_EXPECTED(connect2Low() != -1);
	PPI_EXPECTED(readFromLowPeer() > 0);
	PPI_LOG(OK, currentTestName);

test9:

#undef nextTest
#define nextTest test10
#undef currentTestName
#define currentTestName "Test 9: Cannot open HighI file in write mode, but only in read mode after downgraded (even after socket closed)"
	flushEnv();
	PPI_EXPECTED(openHighI("r+") == NULL);
	PPI_EXPECTED(openHighI("r") != NULL);
	PPI_LOG(OK, currentTestName);

test10:
out:
	flushEnv();
/*
	int fd,rc;

	if (argc > 1) socket_path=argv[1];

	if ( (fd = socket(AF_UNIX, SOCK_STREAM, 0)) == -1) {
		perror("socket error");
		exit(-1);
	}

	FILE *f = fopen("/home/ubuntu/test/lowI", "r");
	if (f) {
		printf("can open lowI for reading\n");
		fgetc(f);
	} else {
		printf("cannot open lowI for reading\n");
	}

	memset(&addr, 0, sizeof(addr));
	addr.sun_family = AF_UNIX;
	strncpy(addr.sun_path, socket_path, sizeof(addr.sun_path)-1);

	if (connect(fd, (struct sockaddr*)&addr, sizeof(addr)) == -1) {
		perror("connect error");
		exit(-1);
	}

	printf("Connected, Please type something ...\n");

	if ( (rc=read(STDIN_FILENO, buf, sizeof(buf))) > 0) {
		if (write(fd, buf, rc) != rc) {
			if (rc > 0) fprintf(stderr,"partial write");
			else {
				perror("write error");
				exit(-1);
			}
		}
	}
	if ( (rc=read(STDIN_FILENO, buf, sizeof(buf))) > 0) {
		if (write(fd, buf, rc) != rc) {
			if (rc > 0) fprintf(stderr,"partial write");
			else {
				perror("write error");
				exit(-1);
			}
		}
	}



	fclose(f);
*/
	return 0;
}
